import copy
import os
import pathlib
import sys
import tempfile
import types
import unittest
from enum import Enum
from unittest.mock import ANY, Mock, mock_open, patch

import git
import numpy as np
import pandas as pd
from funcy import identity

import ballet
import ballet.util
import ballet.util.ci
import ballet.util.fs
import ballet.util.git
import ballet.util.io
from ballet.compat import safepath
from ballet.util.ci import TravisPullRequestBuildDiffer
from ballet.util.code import (  # noqa F401
    blacken_code, get_source, is_valid_python)
from ballet.util.mod import (  # noqa F401
    import_module_at_path, import_module_from_modname,
    import_module_from_relpath, modname_to_relpath, relpath_to_modname)
from ballet.util.testing import ArrayLikeEqualityTestingMixin

from .util import make_mock_commit, make_mock_commits, mock_repo


class UtilTest(
    ArrayLikeEqualityTestingMixin,
    unittest.TestCase
):

    def test_asarray2d_shape_n_x_2(self):
        # case: second dimension is > 1
        a = np.zeros((3, 2))
        result = ballet.util.asarray2d(a)
        self.assertArrayEqual(result, a)

    def test_asarray2d_shape_n_x_1(self):
        # case: second dimension == 1
        a = np.zeros((3, 1))
        result = ballet.util.asarray2d(a)
        self.assertArrayEqual(result, a)

    def test_asarray2d_shape_n(self):
        # case: second dimension not present
        a = np.zeros((3,))
        result = ballet.util.asarray2d(a)
        expected_shape = (3, 1)
        self.assertEqual(result.shape, expected_shape)
        self.assertArrayEqual(np.ravel(result), a)

    def test_asarray2d_series(self):
        # case: pd.Series
        a = np.zeros((3,))
        ser = pd.Series(a)
        result = ballet.util.asarray2d(ser)
        self.assertGreaterEqual(result.shape[1], 1)
        self.assertArrayEqual(
            result, ballet.util.asarray2d(a)
        )

    def test_asarray2d_df(self):
        # case: pd.DataFrame
        a = np.zeros((3, 2))
        df = pd.DataFrame(a)
        result = ballet.util.asarray2d(df)
        self.assertEqual(result.shape, df.shape)
        self.assertGreaterEqual(result.shape[1], 1)
        self.assertArrayEqual(result, a)

    def test_get_arr_desc_array(self):
        a = np.ones((2, 2))
        expected = 'ndarray (2, 2)'
        actual = ballet.util.get_arr_desc(a)
        self.assertEqual(actual, expected)

    def test_get_arr_desc_frame(self):
        df = pd.DataFrame()
        expected = 'DataFrame (0, 0)'
        actual = ballet.util.get_arr_desc(df)
        self.assertEqual(actual, expected)

    def test_get_arr_desc_object(self):
        obj = object()
        expected = 'object <no shape>'
        actual = ballet.util.get_arr_desc(obj)
        self.assertEqual(actual, expected)

    def test_get_enum_keys_class(self):
        class MyEnum:
            A = 1
            B = 2

        actual = ballet.util.get_enum_keys(MyEnum)
        expected = ['A', 'B']
        self.assertEqual(actual, expected)

    def test_get_enum_keys_enum(self):
        class MyEnum(Enum):
            A = 1
            B = 2

        actual = ballet.util.get_enum_keys(MyEnum)
        expected = ['A', 'B']
        self.assertEqual(actual, expected)

    def test_get_enum_values_class(self):
        class MyEnum:
            A = 1
            B = 2

        actual = ballet.util.get_enum_values(MyEnum)
        expected = [1, 2]
        self.assertEqual(actual, expected)

    def test_get_enum_values_enum(self):
        class MyEnum(Enum):
            A = 1
            B = 2

        actual = ballet.util.get_enum_values(MyEnum)
        expected = [1, 2]
        self.assertEqual(actual, expected)

    def test_indent(self):
        text = (
            'Hello\n'
            '  world\n'
            '...hi'
        )
        expected = (
            '    Hello\n'
            '      world\n'
            '    ...hi'
        )
        actual = ballet.util.indent(text, n=4)
        self.assertEqual(actual, expected)

    def test_make_plural_suffix_plural(self):
        objs = ['honda', 'toyota']
        suffix = ballet.util.make_plural_suffix(objs)
        actual = 'car{s}'.format(s=suffix)
        expected = 'cars'
        self.assertEqual(actual, expected)

    def test_make_plural_suffix_singular(self):
        objs = ['honda']
        suffix = ballet.util.make_plural_suffix(objs)
        actual = 'car{s}'.format(s=suffix)
        expected = 'car'
        self.assertEqual(actual, expected)

    def test_whether_failures_failed(self):
        failures = ['Stopped working', 'Made noises']

        @ballet.util.whether_failures
        def do_stuff():
            yield from failures

        success, list_of_failures = do_stuff()
        self.assertFalse(success)
        self.assertEqual(list_of_failures, failures)

    def test_whether_failures_succeeded(self):
        failures = []

        @ballet.util.whether_failures
        def do_stuff():
            yield from failures

        success, list_of_failures = do_stuff()
        self.assertTrue(success)
        self.assertEqual(list_of_failures, failures)

    def test_has_nans(self):
        objs_with_nans = [
            pd.DataFrame(data={'x': [1, np.nan], 'y': [np.nan, 2]}),
            pd.DataFrame(data={'x': [1, np.nan]}),
            pd.Series(data=[1, np.nan]),
            np.array([[1, np.nan], [np.nan, 2]]),
            np.array([[1, np.nan]]),
            np.array([1, np.nan]),
            np.array([1, np.nan]).T,
            np.array(np.nan),
        ]

        objs_without_nans = [
            pd.DataFrame(data={'x': [1, 0], 'y': [0, 2]}),
            pd.DataFrame(data={'x': [1, 0]}),
            pd.Series(data=[1, 0]),
            np.array([[1, 0], [0, 2]]),
            np.array([[1, 0]]),
            np.array([1, 0]),
            np.array([1, 0]).T,
            np.array(0),
        ]

        for obj in objs_with_nans:
            self.assertTrue(ballet.util.has_nans(obj))

        for obj in objs_without_nans:
            self.assertFalse(ballet.util.has_nans(obj))

    def test_dfilter(self):
        @ballet.util.dfilter(lambda x: x >= 0)
        def numbers():
            return [-1, 2, 0, -2]

        actual = numbers()
        expected = [2, 0]
        self.assertEqual(actual, expected)

    def test_load_sklearn_df(self):
        name = 'iris'
        X_df, y_df = ballet.util.load_sklearn_df(name)

        # validation on X_df
        self.assertIsNotNone(X_df)
        self.assertIsInstance(X_df, pd.DataFrame)

        # validation on y_df
        self.assertIsNotNone(y_df)
        self.assertIsInstance(y_df, pd.Series)

    def test_quiet_stderr(self):
        @ballet.util.quiet
        def f():
            print('bar', file=sys.stderr)

        # doesn't check that nothing is printed to sys.stderr, but makes sure
        # the stream is reset properly (esp on py<3.5)
        stderr = sys.stderr
        f()
        self.assertIs(sys.stderr, stderr)

    def test_deepcopy_mixin(self):
        class E(Exception):
            pass

        class A:
            def __init__(self, a):
                self.a = a

            def __deepcopy__(self, memo):
                raise E

        class B(ballet.util.DeepcopyMixin, A):
            pass

        a = A(1)
        with self.assertRaises(E):
            copy.deepcopy(a)

        b = B(1)
        copy.deepcopy(b)

    def test_falsy(self):
        matrix = (
            ('false', True),  # i.e., is falsy
            ('', True),
            ('true', False),  # i.e, is not falsy
            ('123', False),
            (73, False),
        )
        for input, expected in matrix:
            actual = ballet.util.falsy(input)
            self.assertEqual(expected, actual)

            # truthy is complement of falsy
            actual = ballet.util.truthy(input)
            self.assertNotEqual(expected, actual)


class ModTest(unittest.TestCase):

    @unittest.expectedFailure
    def test_import_module_from_modname(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_import_module_from_relpath(self):
        raise NotImplementedError

    def test_import_module_at_path_module(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = pathlib.Path(tmpdir).joinpath('foo', 'bar.py')
            path.parent.mkdir(parents=True)
            init = path.parent.joinpath('__init__.py')
            init.touch()
            x = 1
            with path.open('w') as f:
                f.write('x={x!r}'.format(x=x))
            modname = 'foo.bar'
            modpath = str(path)  # e.g. /tmp/foo/bar.py'
            mod = import_module_at_path(modname, modpath)
            self.assertIsInstance(mod, types.ModuleType)
            self.assertEqual(mod.__name__, modname)
            self.assertEqual(mod.x, x)

        with tempfile.TemporaryDirectory() as tmpdir:
            path = pathlib.Path(tmpdir).joinpath('foo')
            path.mkdir(parents=True)
            init = path.joinpath('__init__.py')
            init.touch()
            x = 'hello'
            with init.open('w') as f:
                f.write('x={x!r}'.format(x=x))
            modname = 'foo'
            modpath = str(path)
            mod = import_module_at_path(modname, modpath)
            self.assertIsInstance(mod, types.ModuleType)
            self.assertEqual(mod.__name__, modname)
            self.assertEqual(mod.x, x)

    @unittest.expectedFailure
    def test_import_module_at_path_bad_package_structure(self):
        raise NotImplementedError

    def test_relpath_to_modname(self):
        relpath = 'ballet/util/_util.py'
        expected_modname = 'ballet.util._util'
        actual_modname = relpath_to_modname(relpath)
        self.assertEqual(actual_modname, expected_modname)

        relpath = 'ballet/util/__init__.py'
        expected_modname = 'ballet.util'
        actual_modname = relpath_to_modname(relpath)
        self.assertEqual(actual_modname, expected_modname)

        relpath = 'ballet/foo/bar/baz.zip'
        with self.assertRaises(ValueError):
            relpath_to_modname(relpath)

    def test_modname_to_relpath(self):
        modname = 'ballet.util._util'
        expected_relpath = 'ballet/util/_util.py'
        actual_relpath = modname_to_relpath(modname)
        self.assertEqual(actual_relpath, expected_relpath)

        modname = 'ballet.util'
        # mypackage.__file__ resolves to the __init__.py
        project_root = pathlib.Path(ballet.__file__).parent.parent

        expected_relpath = 'ballet/util/__init__.py'
        actual_relpath = modname_to_relpath(modname, project_root=project_root)
        self.assertEqual(actual_relpath, expected_relpath)

        # TODO patch this
        # # without providing project root, behavior is undefined, as we don't
        # # know whether the relative path will resolve to a directory

        # # within a temporary directory, the relpath *should not* be a dir
        # with tempfile.TemporaryDirectory() as tmpdir:
        #     cwd = os.getcwd()
        #     try:
        #         os.chdir(tmpdir)
        #         actual_relpath = modname_to_relpath(modname)
        #         expected_relpath = 'ballet/util.py'
        #         self.assertEqual(actual_relpath, expected_relpath)
        #     finally:
        #         os.chdir(cwd)

        # # from the actual project root, the relpath *should* be a dir
        # cwd = os.getcwd()
        # try:
        #     os.chdir(str(project_root))
        #     actual_relpath = modname_to_relpath(modname)
        #     expected_relpath = 'ballet/util/__init__.py'
        #     self.assertEqual(actual_relpath, expected_relpath)
        # finally:
        #     os.chdir(cwd)

        # without add_init
        modname = 'ballet.util'
        add_init = False
        expected_relpath = 'ballet/util'
        actual_relpath = modname_to_relpath(
            modname, project_root=project_root, add_init=add_init)
        self.assertEqual(actual_relpath, expected_relpath)


class CiTest(unittest.TestCase):
    def setUp(self):
        self.pr_num = 7
        self.commit_range = 'HEAD^..HEAD'
        self.travis_vars = {
            'TRAVIS_PULL_REQUEST': str(self.pr_num),
            'TRAVIS_COMMIT_RANGE': self.commit_range,
        }

    def test_get_travis_pr_num(self):
        # matrix of env name, setting for env, expected result
        matrix = (
            ('TRAVIS_PULL_REQUEST', str(self.pr_num), self.pr_num),
            ('TRAVIS_PULL_REQUEST', 'true', None),
            ('TRAVIS_PULL_REQUEST', 'FALSE', None),
            ('TRAVIS_PULL_REQUEST', 'false', None),
            ('TRAVIS_PULL_REQUEST', 'abcd', None),
            ('UNRELATED', '', None),
        )
        for env_name, env_value, expected in matrix:
            with patch.dict('os.environ', {env_name: env_value}, clear=True):
                actual = ballet.util.ci.get_travis_pr_num()
                self.assertEqual(actual, expected)

    def test_is_travis_pr(self):
        matrix = (
            ('TRAVIS_PULL_REQUEST', str(self.pr_num), True),
            ('TRAVIS_PULL_REQUEST', 'true', False),
            ('TRAVIS_PULL_REQUEST', 'FALSE', False),
            ('TRAVIS_PULL_REQUEST', 'false', False),
            ('TRAVIS_PULL_REQUEST', 'abcd', False),
            ('UNRELATED', '', False),
        )
        for env_name, env_value, expected in matrix:
            with patch.dict('os.environ', {env_name: env_value}, clear=True):
                actual = ballet.util.ci.is_travis_pr()
                self.assertEqual(actual, expected)

    def test_get_travis_branch(self):
        # matrix of env dict, expected result
        matrix = (
            ({
                'TRAVIS_PULL_REQUEST': 'false',
                'TRAVIS_PULL_REQUEST_BRANCH': '',
                'TRAVIS_BRANCH': 'master',
            }, 'master'),
            ({
                'TRAVIS_PULL_REQUEST': 'false',
                'TRAVIS_PULL_REQUEST_BRANCH': '',
                'TRAVIS_BRANCH': 'foo',
            }, 'foo'),
            ({
                'TRAVIS_PULL_REQUEST': '1',
                'TRAVIS_PULL_REQUEST_BRANCH': 'foo',
                'TRAVIS_BRANCH': 'master',
            }, 'foo'),
            ({}, None),
        )

        for env, expected in matrix:
            with patch.dict('os.environ', env, clear=True):
                actual = ballet.util.ci.get_travis_branch()
                self.assertEqual(actual, expected)

    def test_travis_pull_request_build_differ(self):
        with mock_repo() as repo:
            make_mock_commits(repo, n=3)
            pr_num = self.pr_num
            commit_range = 'HEAD^..HEAD'

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }
            with patch.dict('os.environ', travis_env_vars, clear=True):
                differ = TravisPullRequestBuildDiffer(pr_num)
                expected_a = repo.rev_parse('HEAD^')
                expected_b = repo.rev_parse('HEAD')
                actual_a, actual_b = differ._get_diff_endpoints()
                self.assertEqual(actual_a, expected_a)
                self.assertEqual(actual_b, expected_b)

    def test_travis_pull_request_build_differ_on_mock_commits(self):
        n = 4
        i = 0
        pr_num = self.pr_num
        feature_branch_name = 'pull/{}'.format(pr_num)
        with mock_repo() as repo:
            make_mock_commit(repo, path='readme.txt')
            expected_merge_base = repo.head.commit
            feature_branch = repo.create_head(feature_branch_name)

            # make commits on branch master
            commits = make_mock_commits(repo, n=3, filename='blah{i}.txt')
            master = repo.heads.master

            # make commits on feature branch
            feature_branch.checkout()
            commits = make_mock_commits(repo, n=n)
            end_commit = commits[-1]

            commit_range = ballet.util.git.make_commit_range(
                master, end_commit)

            travis_env_vars = {
                'TRAVIS_BUILD_DIR': repo.working_tree_dir,
                'TRAVIS_PULL_REQUEST': str(pr_num),
                'TRAVIS_COMMIT_RANGE': commit_range,
            }
            with patch.dict('os.environ', travis_env_vars, clear=True):
                differ = TravisPullRequestBuildDiffer(pr_num)
                a, b = differ._get_diff_endpoints()
                self.assertEqual(a, expected_merge_base)
                self.assertEqual(b, end_commit)

                diffs = differ.diff()

                # there should be n diff objects, they should show files
                # 0 to n-1. merge base just created readme.txt, so all files
                # on feature branch are new.
                self.assertEqual(len(diffs), n)
                j = i
                for diff in diffs:
                    self.assertEqual(diff.change_type, 'A')
                    self.assertEqual(diff.b_path, 'file{j}.py'.format(j=j))
                    j += 1


class FsTest(unittest.TestCase):

    def setUp(self):
        self.fs_path_conversions = [identity, pathlib.Path]

    def test_spliceext(self):
        filepath0 = '/foo/bar/baz.txt'

        for func in self.fs_path_conversions:
            filepath = func(filepath0)
            s = '_new'
            expected = '/foo/bar/baz_new.txt'

            actual = ballet.util.fs.spliceext(filepath, s)
            self.assertEqual(actual, expected)

    def test_replaceext(self):
        filepath0 = '/foo/bar/baz.txt'
        expected = '/foo/bar/baz.py'

        for func in self.fs_path_conversions:
            filepath = func(filepath0)

            new_ext = 'py'
            actual = ballet.util.fs.replaceext(filepath, new_ext)
            self.assertEqual(actual, expected)

            new_ext = '.py'
            actual = ballet.util.fs.replaceext(filepath, new_ext)
            self.assertEqual(actual, expected)

    def test_splitext2(self):
        filepath0 = '/foo/bar/baz.txt'
        expected = ('/foo/bar', 'baz', '.txt')

        for func in self.fs_path_conversions:
            filepath = func(filepath0)

            actual = ballet.util.fs.splitext2(filepath)
            self.assertEqual(actual, expected)

        filepath0 = 'baz.txt'
        expected = ('', 'baz', '.txt')

        for func in self.fs_path_conversions:
            filepath = func(filepath0)

            actual = ballet.util.fs.splitext2(filepath)
            self.assertEqual(actual, expected)

    @patch('os.path.exists')
    def test_isemptyfile_does_not_exist(self, mock_exists):
        mock_exists.return_value = False
        result = ballet.util.fs.isemptyfile(
            '/path/to/file/that/hopefully/does/not/exist')
        self.assertFalse(result)

    def test_isemptyfile_is_not_empty(self):
        # file exists and is not empty - false
        with tempfile.TemporaryDirectory() as d:
            filepath = os.path.join(d, 'file')
            with open(filepath, 'w') as f:
                f.write('0')
            result = ballet.util.fs.isemptyfile(filepath)
            self.assertFalse(result)

    def test_isemptyfile_is_empty(self):
        # file exists and is empty - true
        with tempfile.TemporaryDirectory() as d:
            filepath = pathlib.Path(d).joinpath('file')
            filepath.touch()
            result = ballet.util.fs.isemptyfile(filepath)
            self.assertTrue(result)

    def test_synctree(self):
        with tempfile.TemporaryDirectory() as tempdir:
            tempdir = pathlib.Path(tempdir).resolve()

            src = tempdir.joinpath('x')
            src.joinpath('a', 'b').mkdir(parents=True)
            src.joinpath('a', 'b', 'only_in_src.txt').touch()
            src.joinpath('a', 'c').mkdir()

            dst = tempdir.joinpath('y')
            dst.joinpath('a', 'b').mkdir(parents=True)
            dst.joinpath('a', 'only_in_dst.txt').touch()

            # patch here in order to avoid messing up tempdir stuff
            with patch('pathlib.Path.mkdir') as mock_mkdir, \
                    patch('os.unlink') as mock_unlink, \
                    patch('os.rmdir') as mock_rmdir, \
                    patch('ballet.util.fs.copyfile') as mock_copyfile:
                ballet.util.fs.synctree(src, dst)

        # one call to mkdir, for 'a/c'
        mock_mkdir.assert_called_once_with()

        # one call to copyfile, for 'only_in_src.txt'
        path = ('a', 'b', 'only_in_src.txt')
        mock_copyfile.assert_called_once_with(
            src.joinpath(*path),
            dst.joinpath(*path)
        )

        # no calls to cleanup
        mock_rmdir.assert_not_called()
        mock_unlink.assert_not_called()

    @unittest.skip
    def test__synctree(self):
        # when src is a directory that exists and dst does not exist,
        # then copytree should be called
        src = Mock(spec=pathlib.Path)
        dst = Mock(spec=pathlib.Path)
        dst.exists.return_value = False
        ballet.util.fs._synctree(src, dst, lambda x: None)

    @unittest.expectedFailure
    def test_pwalk(self):
        raise NotImplementedError


class GitTest(unittest.TestCase):

    def test_make_commit_range(self):
        a = 'abc1234'
        b = 'def4321'
        expected_commit_range = 'abc1234...def4321'
        actual_commit_range = ballet.util.git.make_commit_range(a, b)
        self.assertEqual(actual_commit_range, expected_commit_range)

    @unittest.expectedFailure
    def test_get_diff_endpoints_from_commit_range(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_get_repo(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_get_pr_num(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_get_branch(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_switch_to_new_branch(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_set_config_variables(self):
        raise NotImplementedError

    @patch('requests.get')
    def test_get_pull_requests(self, mock_requests_get):
        owner = 'foo'
        repo = 'bar'
        state = 'closed'
        ballet.util.git.get_pull_requests(owner, repo, state=state)

        (url, ), kwargs = mock_requests_get.call_args
        self.assertIn(owner, url)
        self.assertIn(repo, url)
        self.assertIn('headers', kwargs)
        self.assertIn('params', kwargs)
        self.assertEqual(kwargs['params']['state'], state)

    @patch('ballet.util.git.get_pull_requests')
    def test_get_pull_request_outcomes(self, mock_get_pull_requests):
        mock_get_pull_requests.return_value = [
            {
                'id': 1,
                "created_at": "2011-01-26T19:01:12Z",
                "updated_at": "2011-01-26T19:01:12Z",
                "closed_at": "2011-01-26T19:01:12Z",
                "merged_at": "2011-01-26T19:01:12Z",
            },
            {
                "created_at": "2011-01-26T19:03:19Z",
                "updated_at": "2011-01-26T19:03:19Z",
                "closed_at": "2011-01-26T19:04:01Z",
                "merged_at": None,
            }
        ]
        owner = 'foo'
        repo = 'bar'

        expected = ['accepted', 'rejected']
        actual = ballet.util.git.get_pull_request_outcomes(owner, repo)
        self.assertEqual(actual, expected)
        mock_get_pull_requests.assert_called_once_with(
            owner, repo, state='closed')

    def test_did_git_push_succeed(self):
        local_ref = None
        remote_ref_string = None
        remote = None

        flags = 0
        push_info = git.remote.PushInfo(flags, local_ref, remote_ref_string,
                                        remote)
        self.assertTrue(ballet.util.git.did_git_push_succeed(push_info))

        flags = git.remote.PushInfo.ERROR
        push_info = git.remote.PushInfo(flags, local_ref, remote_ref_string,
                                        remote)
        self.assertFalse(ballet.util.git.did_git_push_succeed(push_info))


class IoTest(unittest.TestCase):

    def setUp(self):
        self.array = np.arange(10).reshape(2, 5)
        self.frame = pd.util.testing.makeDataFrame()

    def test_check_ext_valid(self):
        ext = '.py'
        expected = ext
        ballet.util.io._check_ext(ext, expected)

    def test_check_ext_invalid_throws(self):
        ext = '.py'
        expected = '.txt'
        with self.assertRaises(ValueError):
            ballet.util.io._check_ext(ext, expected)

    @patch('ballet.util.io._write_tabular_pickle')
    @patch('ballet.util.io._write_tabular_h5')
    def test_write_tabular(self, mock_write_tabular_h5,
                           mock_write_tabular_pickle):
        obj = object()
        filepath = '/foo/bar/baz.h5'
        ballet.util.io.write_tabular(obj, filepath)
        mock_write_tabular_h5.assert_called_once_with(obj, filepath)

        obj = object()
        filepath = '/foo/bar/baz.pkl'
        ballet.util.io.write_tabular(obj, filepath)
        mock_write_tabular_pickle.assert_called_once_with(obj, filepath)

        obj = object()
        filepath = '/foo/bar/baz.xyz'
        with self.assertRaises(NotImplementedError):
            ballet.util.io.write_tabular(obj, filepath)

    @patch('builtins.open', new_callable=mock_open)
    @patch('pickle.dump')
    def test_write_tabular_pickle_ndarray(self, mock_dump, mock_open):
        obj = self.array
        filepath = '/foo/bar/baz.pkl'
        ballet.util.io._write_tabular_pickle(obj, filepath)
        mock_dump.assert_called_with(obj, ANY)

    @patch('builtins.open', new_callable=mock_open)
    def test_write_tabular_pickle_ndframe(self, mock_open):
        obj = self.frame
        filepath = '/foo/bar/baz.pkl'

        with patch.object(obj, 'to_pickle') as mock_to_pickle:
            ballet.util.io._write_tabular_pickle(obj, filepath)

        mock_to_pickle.assert_called_with(filepath)

    def test_write_tabular_pickle_nonarray_raises(self):
        obj = object()
        filepath = '/foo/bar/baz.pkl'
        with self.assertRaises(NotImplementedError):
            ballet.util.io._write_tabular_pickle(obj, filepath)

    def test_write_tabular_h5_ndarray(self):
        obj = self.array
        with tempfile.TemporaryDirectory() as d:
            filepath = pathlib.Path(d).joinpath('baz.h5')
            ballet.util.io._write_tabular_h5(obj, filepath)

            file_size = os.path.getsize(safepath(filepath))
            self.assertGreater(file_size, 0)

    def test_write_tabular_h5_ndframe(self):
        obj = self.frame
        filepath = '/foo/bar/baz.h5'

        with patch.object(obj, 'to_hdf') as mock_to_hdf:
            ballet.util.io._write_tabular_h5(obj, filepath)

        mock_to_hdf.assert_called_with(filepath, key=ANY)

    @unittest.expectedFailure
    def test_read_tabular(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_save_model(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_save_predictions(self):
        raise NotImplementedError


class CodeTest(unittest.TestCase):

    def test_is_valid_python(self):
        code = '1'
        result = is_valid_python(code)
        self.assertTrue(result)

    def test_is_valid_python_invalid(self):
        code = 'this is not valid python code'
        result = is_valid_python(code)
        self.assertFalse(result)

    @unittest.skipUnless(sys.version_info >= (3, 6),
                         "black requires py36 or higher")
    def test_blacken_code(self):
        input = '''\
        x = {  'a':37,'b':42,

        'c':927}
        '''.strip()

        expected = 'x = {"a": 37, "b": 42, "c": 927}'.strip()
        actual = blacken_code(input).strip()

        self.assertEqual(actual, expected)

    @unittest.skipUnless(sys.version_info >= (3, 6),
                         "black requires py36 or higher")
    def test_blacken_code_nothing_changed(self):
        input = '1\n'
        expected = '1\n'
        actual = blacken_code(input)

        self.assertEqual(actual, expected)

    @unittest.expectedFailure
    def test_get_source(self):
        raise NotImplementedError



import unittest
from unittest.mock import ANY, patch

from click.testing import CliRunner

import ballet
from ballet.cli import (  # noqa F401
    cli, quickstart, start_new_feature, update_project_template, validate)


class CliTest(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    def test_cli_version(self):
        result = self.runner.invoke(cli, '--version')
        self.assertEqual(result.exit_code, 0)
        self.assertIn(ballet.__version__, result.output)

    @patch('ballet.templating.render_project_template')
    def test_quickstart(self, mock_render):
        result = self.runner.invoke(quickstart)
        mock_render.assert_called_once_with()
        self.assertIn('Generating new ballet project', result.output)

    @patch('ballet.update.update_project_template')
    def test_update_project_template(self, mock_update):
        result = self.runner.invoke(update_project_template, ['--push'])
        mock_update.assert_called_once_with(
            push=True, project_template_path=None)
        self.assertIn('Updating project template', result.output)

    @patch('ballet.templating.start_new_feature')
    def test_start_new_feature(self, mock_start):
        result = self.runner.invoke(start_new_feature)
        mock_start.assert_called_once_with()
        self.assertIn('Starting new feature', result.output)

    @patch('ballet.project.Project.from_path')
    @patch('ballet.validation.main.validate')
    def test_validate_all(self, mock_validate, mock_from_path):
        result = self.runner.invoke(validate, ['-A'])
        mock_validate.assert_called_once_with(ANY, True, True, True, True)
        self.assertIn('Validating project', result.output)



from funcy import decorator, ignore
from stacklog import stacklog

from ballet.exc import (
    FeatureRejected, InvalidFeatureApi, InvalidProjectStructure,
    NoFeaturesCollectedError, SkippedValidationTest)
from ballet.util.log import logger
from ballet.util.mod import import_module_from_modname
from ballet.validation.common import (
    get_accepted_features, get_proposed_feature)

# helpful for log parsing
PRUNER_MESSAGE = 'Found Redundant Feature: '


@decorator
def validation_stage(call, message):
    call = stacklog(logger.info,
                    'Ballet Validation: {message}'.format(message=message),
                    conditions=[(SkippedValidationTest, 'SKIPPED')])(call)
    call = ignore(SkippedValidationTest)(call)
    return call()


def _load_class(project, config_key):
    path = project.config.get(config_key)
    modname, clsname = path.rsplit('.', maxsplit=1)
    mod = import_module_from_modname(modname)
    cls = getattr(mod, clsname)

    clsname = getattr(cls, '__name__', '<unknown>')
    modfile = getattr(mod, '__file__', '<unknown>')
    logger.debug('Loaded class {} from {}'.format(clsname, modfile))

    return cls


@validation_stage('checking project structure')
def _check_project_structure(project, force=False):
    if not force and not project.on_pr:
        raise SkippedValidationTest('Not on PR')

    Validator = _load_class(project, 'validation.project_structure_validator')
    validator = Validator(project)
    result = validator.validate()
    if not result:
        raise InvalidProjectStructure


@validation_stage('validating feature API')
def _validate_feature_api(project, force=False):
    """Validate feature API"""
    if not force and not project.on_pr:
        raise SkippedValidationTest('Not on PR')

    Validator = _load_class(project, 'validation.feature_api_validator')
    validator = Validator(project)
    result = validator.validate()
    if not result:
        raise InvalidFeatureApi


@validation_stage('evaluating feature performance')
def _evaluate_feature_performance(project, force=False):
    """Evaluate feature performance"""
    if not force and not project.on_pr:
        raise SkippedValidationTest('Not on PR')

    out = project.build()
    X_df, y, features = out.X_df, out.y, out.features

    proposed_feature = get_proposed_feature(project)
    accepted_features = get_accepted_features(features, proposed_feature)

    Accepter = _load_class(project, 'validation.feature_accepter')
    accepter = Accepter(X_df, y, accepted_features, proposed_feature)
    accepted = accepter.judge()

    if not accepted:
        raise FeatureRejected


@validation_stage('pruning existing features')
def _prune_existing_features(project, force=False):
    """Prune existing features"""
    if not force and not project.on_master_after_merge:
        raise SkippedValidationTest('Not on master')

    try:
        proposed_feature = get_proposed_feature(project)
    except NoFeaturesCollectedError:
        raise SkippedValidationTest('No features collected')

    out = project.build()
    X_df, y, features = out.X_df, out.y, out.features
    accepted_features = get_accepted_features(features, proposed_feature)

    Pruner = _load_class(project, 'validation.feature_pruner')
    pruner = Pruner(X_df, y, accepted_features, proposed_feature)
    redundant_features = pruner.prune()

    # "propose removal"
    for feature in redundant_features:
        logger.info(PRUNER_MESSAGE + feature.source)

    return redundant_features


def validate(project,
             check_project_structure,
             check_feature_api,
             evaluate_feature_acceptance,
             evaluate_feature_pruning):
    """Entrypoint for 'ballet validate' command in ballet projects"""
    if check_project_structure:
        _check_project_structure(project)
    if check_feature_api:
        _validate_feature_api(project)
    if evaluate_feature_acceptance:
        _evaluate_feature_performance(project)
    if evaluate_feature_pruning:
        _prune_existing_features(project)



import numpy as np

from ballet.util import asarray2d

LAMBDA_1_ADJUSTMENT = 64
LAMBDA_2_ADJUSTMENT = 64


def _concat_datasets(dfs_by_src, n_samples=0, omit=None):
    if omit is None:
        omit = []
    filtered_dfs = [np.array(dfs_by_src[x])
                    for x in dfs_by_src if x not in omit]
    if len(filtered_dfs) == 0:
        return np.zeros((n_samples, 1))
    return asarray2d(np.concatenate(filtered_dfs, axis=1))


def _compute_lmbdas(unnorm_lmbda_1, unnorm_lmbda_2, features_by_src):
    feat_srcs = features_by_src.keys()
    num_features = len(feat_srcs)
    num_feature_cols = 0
    for feat_src in features_by_src:
        num_feature_cols += features_by_src[feat_src].shape[1]
    return (unnorm_lmbda_1 / num_features, unnorm_lmbda_2 / num_feature_cols)


def _compute_threshold(lmbda_1, lmbda_2, n_feature_cols, n_omitted_cols=0):
    return lmbda_1 + lmbda_2 * (n_feature_cols - n_omitted_cols)



import unittest
from unittest.mock import MagicMock, patch

import numpy as np

from ballet import Feature
from ballet.eng.base import SimpleFunctionTransformer
from ballet.eng.misc import IdentityTransformer
from ballet.util import asarray2d
from ballet.validation.feature_pruning.validator import (
    GFSSFPruner, NoOpPruner, RandomPruner)
from tests.util import load_regression_data


class NoOpPrunerTest(unittest.TestCase):

    def test_pruner(self):
        X = None
        y = None
        existing_features = []
        feature = None

        expected = []

        pruner = NoOpPruner(X, y, existing_features, feature)
        actual = pruner.prune()

        self.assertEqual(expected, actual)


class RandomPrunerTest(unittest.TestCase):

    @patch('random.choice')
    @patch('random.uniform')
    def test_pruner(self, mock_uniform, mock_choice):
        existing_features = [MagicMock(), MagicMock()]

        mock_uniform.return_value = 0.0  # makes sure pruning is performed
        mock_choice.return_value = existing_features[0]  # the pruned feature
        expected = [existing_features[0]]

        X = None
        y = None
        feature = None

        pruner = RandomPruner(X, y, existing_features, feature)
        actual = pruner.prune()

        self.assertEqual(expected, actual)


class GFSSFPrunerTest(unittest.TestCase):

    def setUp(self):
        self.X, self.y = load_regression_data(n_informative=1,
                                              n_uninformative=14)

    def test_prune_exact_replicas(self):
        feature_1 = Feature(
            input='A_0',
            transformer=IdentityTransformer(),
            source='1st Feature')
        feature_2 = Feature(
            input='A_0',
            transformer=IdentityTransformer(),
            source='2nd Feature')
        gfssf_pruner = GFSSFPruner(
            self.X, self.y, [feature_1], feature_2)

        redunant_features = gfssf_pruner.prune()
        self.assertIn(
            feature_1,
            redunant_features,
            'Exact replica features should be pruned')

    @unittest.expectedFailure
    def test_prune_weak_replicas(self):
        def add_noise(X):
            X = asarray2d(X)
            return X + np.random.normal(0, 0.5, X.shape)

        noise_transformer = SimpleFunctionTransformer(add_noise)
        feature_weak = Feature(
            input='A_0',
            transformer=noise_transformer,
            source='1st Feature')
        feature_strong = Feature(
            input='A_0',
            transformer=IdentityTransformer(),
            source='2nd Feature')
        gfssf_pruner = GFSSFPruner(
            self.X, self.y, [feature_weak], feature_strong)

        redunant_features = gfssf_pruner.prune()
        self.assertIn(
            feature_weak,
            redunant_features,
            'Noisy features should be pruned')

    def test_prune_keep_relevant(self):
        feature_1 = Feature(
            input='A_0',
            transformer=IdentityTransformer(),
            source='1st Feature')
        feature_2 = Feature(
            input='Z_0',
            transformer=IdentityTransformer(),
            source='2nd Feature')
        gfssf_pruner = GFSSFPruner(
            self.X, self.y, [feature_1], feature_2)

        redunant_features = gfssf_pruner.prune()
        self.assertNotIn(
            feature_1,
            redunant_features,
            'Still relevant features should be pruned')

    @unittest.expectedFailure
    def test_prune_irrelevant_features(self):
        feature_1 = Feature(
            input='Z_0',
            transformer=IdentityTransformer(),
            source='1st Feature')
        feature_2 = Feature(
            input='A_0',
            transformer=IdentityTransformer(),
            source='2nd Feature')
        gfssf_pruner = GFSSFPruner(
            self.X, self.y, [feature_1], feature_2)

        redunant_features = gfssf_pruner.prune()
        self.assertIn(
            feature_1,
            redunant_features,
            'Irrelevant features should be pruned')



import pathlib
import random
import subprocess
import tempfile

import git
from funcy import any_fn, contextmanager, ignore, merge
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn_pandas.pipeline import TransformerPipeline

from ballet.eng.misc import IdentityTransformer
from ballet.util.git import set_config_variables
from ballet.util.log import logger

EPSILON = 1e-4


class FragileTransformer(BaseEstimator, TransformerMixin):

    def __init__(self, bad_input_checks, errors):
        """Raises a random error if any input check returns True"""
        super().__init__()

        self._check = any_fn(*bad_input_checks)
        self._errors = errors

        self._random = random.Random()
        self._random.seed = hash(merge(bad_input_checks, errors))

    def _raise(self):
        raise self._random.choice(self._errors)

    def fit(self, X, y=None, **fit_kwargs):
        if self._check(X) or self._check(y):
            self._raise()

        return self

    def transform(self, X, **transform_kwargs):
        if self._check(X):
            self._raise()

        return X


class FragileTransformerPipeline(TransformerPipeline):
    def __init__(self, nsteps, bad_input_checks, errors, shuffle=True, seed=1):
        steps = [
            ('IdentityTransformer{:02d}'.format(i), IdentityTransformer())
            for i in range(nsteps - 1)
        ]
        fragile_transformer = FragileTransformer(bad_input_checks, errors)
        steps.append(
            (repr(fragile_transformer), fragile_transformer)
        )
        if shuffle:
            rand = random.Random()
            rand.seed(seed)
            rand.shuffle(steps)

        super().__init__(steps)


def make_mock_commit(repo, kind='A', path=None, content=None):
    """Commits one file to repo"""
    if not path:
        path = 'file{}'.format(random.randint(0, 999))

    dir = repo.working_tree_dir
    abspath = pathlib.Path(dir).joinpath(path)

    if kind == 'A':
        # TODO make robust
        abspath.parent.mkdir(parents=True, exist_ok=True)

        if abspath.exists():
            # because this would be a kind=='M'
            raise FileExistsError(str(abspath))
        else:
            if content is not None:
                with abspath.open('w') as f:
                    f.write(content)
            else:
                abspath.touch()
        repo.git.add(str(abspath))
        repo.git.commit(m='Commit {}'.format(str(abspath)))
    else:
        raise NotImplementedError

    return repo.head.commit


def make_mock_commits(repo, n=10, filename='file{i}.py'):
    """Create n sequential files/commits"""
    if '{i}' not in filename:
        raise ValueError

    commits = []
    for i in range(n):
        commit = make_mock_commit(repo, path=filename.format(i=i))
        commits.append(commit)
    return commits


def set_ci_git_config_variables(repo, name='Foo Bar', email='foo@bar.com'):
    set_config_variables(repo, {
        'user.name': name,
        'user.email': email,
    })


@contextmanager
def mock_repo():
    """Create a new repo"""
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = pathlib.Path(tempdir)
        repo = git.Repo.init(str(tempdir))
        set_ci_git_config_variables(repo)
        yield repo


@ignore((FileNotFoundError, subprocess.CalledProcessError))
def tree(dir):
    cmd = ['tree', '-A', '-n', '--charset', 'ASCII', str(dir)]
    logger.debug('Popen({cmd!r})'.format(cmd=cmd))
    tree_output = subprocess.check_output(cmd).decode()
    logger.debug(tree_output)
    return tree_output


def load_regression_data(n_informative=1, n_uninformative=14, n_samples=500):
    import pandas as pd
    from sklearn.datasets import make_regression

    q = n_informative
    p = n_informative + n_uninformative
    X, y, coef = make_regression(
        n_samples=n_samples, n_features=p, n_informative=q, coef=True,
        shuffle=True, random_state=1)

    # informative columns are 'A_0', 'A_1', ...
    informative = ['A_{i}'.format(i=i) for i in range(n_informative)]

    # uninformative columns are 'Z_0', 'Z_1', ...
    uninformative = ['Z_{i}'.format(i=i) for i in range(n_uninformative)]

    columns = [
        informative.pop(0) if c != 0 else uninformative.pop(0)
        for c in coef
    ]

    X_df = pd.DataFrame(data=X, columns=columns)
    y_df = pd.Series(y)
    return X_df, y_df



import collections

from funcy import iterable
from sklearn_pandas import DataFrameMapper

import ballet.feature
from ballet.eng.misc import NullTransformer


class FeatureEngineeringPipeline(DataFrameMapper):
    """Feature engineering pipeline

    Args:
        features (Union[Feature, List[Feature]]): feature or list of features
    """

    def __init__(self, features):
        if not features:
            features = ballet.feature.Feature(input=[],
                                              transformer=NullTransformer())

        if not iterable(features):
            features = (features, )

        super().__init__(
            [t.as_input_transformer_tuple() for t in features],
            input_df=True)


BuildResult = collections.namedtuple(
    'BuildResult',
    ['X_df', 'features', 'mapper_X', 'X', 'y_df', 'encoder_y', 'y'])



import logging

import ballet

SIMPLE_LOG_FORMAT = r'%(asctime)s %(levelname)s - %(message)s'
DETAIL_LOG_FORMAT = r'[%(asctime)s] {%(name)s: %(filename)s:%(lineno)d} %(levelname)s - %(message)s'  # noqa E501

logger = logging.getLogger(ballet.__name__)
_handler = None


def enable(logger=logger,
           level=logging.INFO,
           format=DETAIL_LOG_FORMAT,
           echo=True):
    """Enable simple console logging for this module

    Args:
        logger (logging.Logger): logger to enable. Defaults to ballet logger.
        level (str, int): logging level, either as string (``"INFO"``) or as
            int (``logging.INFO`` or ``20``). Defaults to 'INFO'.
        format (str) : logging format. Defaults to
            ballet.util.log.DETAIL_LOG_FORMAT.
        echo (bool): Whether to log a message at the configured log level to
            confirm that logging is enable. Defaults to True.
    """
    global _handler
    if _handler is None:
        _handler = logging.StreamHandler()
        formatter = logging.Formatter(format)
        _handler.setFormatter(formatter)

    level = logging._checkLevel(level)
    levelName = logging._levelToName[level]

    logger.setLevel(level)
    _handler.setLevel(level)

    if _handler not in logger.handlers:
        logger.addHandler(_handler)

    if echo:
        logger.log(
            level, 'Logging enabled at level {name}.'.format(name=levelName))


class LevelFilter(object):
    """Logging filter for log records at an exact level

    Args:
        level (int): numeric logging level to filter

    Example usage:

        >>> debugFilter = LevelFilter(logging.DEBUG)
        >>> logger.addFilter(debugFilter)
    """

    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno == self.__level


class LoggingContext(object):
    """Logging context manager
    
    Configure the given logger to emit messages at and above the configured 
    logging level and using the configured logging handler. Useful to 
    temporarily set a lower (or higher log level) or to temporarily add a 
    local handler. After the context exits, the original state of the logger
    will be restored.
    
    Args:
        logger (logging.Logger): logger
        level (Union[str,int]): string or numeric logging level
        handler (logging.Handler): log handler if not is already configured
        close (bool): whether to close the handler after context exits. 
            Defaults to True.
            
    Example usage:
    
        >>> with LoggingContext(logger, level='DEBUG'):
        ...     logger.debug('some message')
        
    Source: <https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging>
    """  # noqa E501

    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler is not None:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler is not None:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        return None



import unittest

import numpy as np

from ballet.compat import SimpleImputer
from ballet.eng.base import BaseTransformer
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.util import has_nans
from ballet.validation.common import check_from_class
from ballet.validation.feature_api.checks import (
    CanDeepcopyCheck, CanTransformCheck, FeatureApiCheck,
    HasCorrectInputTypeCheck, HasCorrectOutputDimensionsCheck,
    NoMissingValuesCheck)

from .util import SampleDataMixin
from ..util import FragileTransformer


class FeatureApiCheckTest(SampleDataMixin, unittest.TestCase):

    def test_good_feature(self):
        feature = Feature(
            input='size',
            transformer=SimpleImputer(),
        )

        valid, failures = check_from_class(
            FeatureApiCheck, feature, self.X, self.y)
        self.assertTrue(valid)
        self.assertEqual(len(failures), 0)

    def test_bad_feature_input(self):
        # bad input
        feature = Feature(
            input=3,
            transformer=SimpleImputer(),
        )
        valid, failures = check_from_class(
            FeatureApiCheck, feature, self.X, self.y)
        self.assertFalse(valid)
        self.assertIn(HasCorrectInputTypeCheck.__name__, failures)

    def test_bad_feature_transform_errors(self):
        # transformer throws errors
        feature = Feature(
            input='size',
            transformer=FragileTransformer(
                (lambda x: True, ), (RuntimeError, ))
        )
        valid, failures = check_from_class(
            FeatureApiCheck, feature, self.X, self.y)
        self.assertFalse(valid)
        self.assertIn(CanTransformCheck.__name__, failures)

    def test_bad_feature_wrong_transform_length(self):
        class _WrongLengthTransformer(BaseTransformer):
            def transform(self, X, **transform_kwargs):
                new_shape = list(X.shape)
                new_shape[0] += 1
                output = np.arange(np.prod(new_shape)).reshape(new_shape)
                return output

        # doesn't return correct length
        feature = Feature(
            input='size',
            transformer=_WrongLengthTransformer(),
        )
        valid, failures = check_from_class(
            FeatureApiCheck, feature, self.X, self.y)
        self.assertFalse(valid)
        self.assertIn(HasCorrectOutputDimensionsCheck.__name__, failures)

    def test_bad_feature_deepcopy_fails(self):
        class _CopyFailsTransformer(IdentityTransformer):
            def __deepcopy__(self, memo):
                raise RuntimeError
        feature = Feature(
            input='size',
            transformer=_CopyFailsTransformer(),
        )
        valid, failures = check_from_class(
            FeatureApiCheck, feature, self.X, self.y)
        self.assertFalse(valid)
        self.assertIn(CanDeepcopyCheck.__name__, failures)

    def test_producing_missing_values_fails(self):
        assert has_nans(self.X)
        feature = Feature(
            input='size',
            transformer=IdentityTransformer()
        )
        valid, failures = check_from_class(
            FeatureApiCheck, feature, self.X, self.y)
        self.assertFalse(valid)
        self.assertIn(NoMissingValuesCheck.__name__, failures)



import unittest
from unittest.mock import ANY, PropertyMock, patch

from ballet.project import Project, detect_github_username


class ProjectTest(unittest.TestCase):

    @unittest.expectedFailure
    def test_load_config_at_path(self):
        raise NotImplementedError

    @unittest.expectedFailure
    def test_load_config_in_dir(self):
        raise NotImplementedError

    @patch('ballet.project.Project.repo', new_callable=PropertyMock)
    def test_detect_github_username_config(self, mock_project_repo):
        expected_username = 'Foo Bar'

        # output of project.repo.config_reader().get_value(...)
        mock_get_value = (mock_project_repo
                          .return_value
                          .config_reader
                          .return_value
                          .get_value)
        mock_get_value.return_value = expected_username

        project = Project(None)
        username = detect_github_username(project)
        self.assertEqual(expected_username, username)

        mock_get_value.assert_called_with('github', 'user', default=ANY)

    @patch('ballet.project.Project.repo', new_callable=PropertyMock)
    @patch('ballet.project.get_pr_num')
    def test_project_pr_num(self, mock_get_pr_num, mock_repo):
        expected = 3
        mock_get_pr_num.return_value = expected

        package = None
        project = Project(package)
        self.assertEqual(project.pr_num, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

requirements = [
    'black; python_version>="3.6"',
    'cookiecutter',
    'Click>=6.0',
    'dill',
    'dynaconf',
    'funcy',
    'gitpython',
    'h5py',
    'numpy',
    'packaging',
    'pandas',
    'pyyaml',
    'requests',
    'scikit_learn>=0.20',
    'scipy',
    'sklearn_pandas',
    'stacklog',
]

setup_requirements = [
    'pytest-runner>=2.11.1',
]

test_requirements = [
    'coverage>=4.5.1',
    'pytest>=3.4.2',
    'pytest-cov>=2.6',
    'pytest-virtualenv>=1.7.0',
    'tox>=2.9.1',
]

development_requirements = [
    # general
    'bumpversion>=0.5.3',
    'pip>=9.0.1',
    'watchdog>=0.8.3',

    # docs
    'm2r>=0.2.0',
    'Sphinx>=1.7.1',
    'sphinx_rtd_theme>=0.2.4',
    'sphinx-click>=1.4.1',

    # style check
    'flake8>=3.5.0',
    'isort>=4.3.4,<=4.3.9',

    # fix style issues
    'autopep8>=1.3.5',

    # distribute on PyPI
    'twine>=1.10.0',
    'wheel>=0.30.0',
]

setup(
    author="Micah Smith",
    author_email='micahs@mit.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description='Core functionality for lightweight, collaborative data science projects',
    entry_points = {
        'console_scripts': [
            'ballet=ballet.cli:cli',
        ],
    },
    extras_require={
        'test': test_requirements,
        'dev': development_requirements + test_requirements,
    },
    install_requires=requirements,
    license='MIT license',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='ballet',
    name='ballet',
    packages=find_packages(include=['ballet', 'ballet.*']),
    python_requires='>=3.5',
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/HDI-Project/ballet',
    version='0.6.10',
    zip_safe=False,
)



from ballet import logger
from ballet.validation.common import (
    check_from_class, subsample_data_for_validation)
from ballet.validation.feature_api.checks import FeatureApiCheck


def validate_feature_api(feature, X, y, subsample=False):
    logger.debug('Validating feature {feature!r}'.format(feature=feature))
    if subsample:
        X, y = subsample_data_for_validation(X, y)
    valid, failures = check_from_class(FeatureApiCheck, feature, X, y)
    if valid:
        logger.info('Feature is valid')
    else:
        logger.info(
            'Feature is NOT valid; failures were {failures}'
            .format(failures=failures))
    return valid



import json
import pathlib
import subprocess
import sys
import tempfile
from textwrap import dedent
from unittest.mock import patch

import funcy
import git
import packaging.version
from cookiecutter.prompt import prompt_for_config
from funcy import complement, lfilter, re_find, re_test
from git import GitCommandError
from stacklog import stacklog

import ballet
from ballet.compat import safepath
from ballet.exc import BalletError, ConfigurationError
from ballet.project import Project
from ballet.templating import render_project_template
from ballet.util.git import did_git_push_succeed
from ballet.util.log import logger

PROJECT_CONTEXT_PATH = (
    pathlib.Path(__file__).resolve().parent.joinpath(
        'templates',
        'project_template',
        'cookiecutter.json'))
CONTEXT_FILE_NAME = '.cookiecutter_context.json'
DEFAULT_BRANCH = 'master'
TEMPLATE_BRANCH = 'project-template'


def _query_pip_search_ballet():
    """Call python -m pip search ballet"""
    # compat: use subprocess.run on py37+ and text=True on py37+
    popen_args = [sys.executable, '-m', 'pip', 'search', 'ballet']
    return subprocess.check_output(popen_args, universal_newlines=True)


def _extract_latest_from_search_triple(triple):
    """Try to extract latest version number from a triple of search results"""
    description, installed, latest = triple
    if re_test(r'\s*ballet \(.+\)\s*-\s*\w*', description):
        if 'INSTALLED' in installed and 'LATEST' in latest:
            return re_find(r'\s*LATEST:\s*(.+)', latest)
    return None


def _get_latest_ballet_version_string():
    """Get the latest version of ballet according to pip

    Parses the result of `pip search ballet`. Looks for something named
    `ballet` to be installed. If something appears to be named `ballet` but
    is not installed, then obviously it is not correct because otherwise how
    would this code be running? :)

    Returns:
        Union[str, None]: latest version of ballet or None if something went
            wrong
    """

    # $ pip search ballet
    # something-else-that-has-ballet-in-the-name (1.1.1)  - some description
    # ballet (x.y.z)  - some description
    #   INSTALLED: x.y.z
    #   LATEST:    u.v.w
    # something-else-that-has-ballet-in-the-name (1.1.1)  - some description

    output = _query_pip_search_ballet()
    for triple in funcy.partition(3, 1, output):
        match = _extract_latest_from_search_triple(triple)
        if match:
            return match

    return None


def _check_for_updated_ballet():
    """Return the version of an updated ballet if it is available

    Returns:
        Union[str, None]: the latest version of ballet available or None if
            the latest version is the same as the currently-installed version
    """
    latest = _get_latest_ballet_version_string()
    current = ballet.__version__
    parse = packaging.version.parse
    if latest and parse(latest) > parse(current):
        return latest
    else:
        return None


def _warn_of_updated_ballet(latest):
    if latest is not None:
        msg = '''\
                A new version of ballet is available: {latest}
                - you currently have ballet v{current}
                - if you don't update ballet, you won't receive project template updates
                - update ballet and then try again:
                
                    $ pip install --upgrade ballet'
                '''  # noqa E501
        msg = msg.format(latest=latest, current=ballet.__version__)
        msg = dedent(msg)
        logger.warn(msg)


def _make_template_branch_merge_commit_message():
    version = ballet.__version__
    return 'Merge project template updates from ballet v{}'.format(version)


def _safe_delete_remote(repo, name):
    with funcy.suppress(Exception):
        repo.delete_remote(name)


def _render_project_template(cwd, tempdir, project_template_path=None):
    tempdir = pathlib.Path(tempdir)
    context = _get_full_context(cwd)

    # don't dump replay files to home directory.
    with patch('cookiecutter.main.dump'):
        return render_project_template(
            project_template_path=project_template_path,
            no_input=True,
            extra_context=context,
            output_dir=safepath(tempdir))


def _get_full_context(cwd):
    # load the context stored within the project repository
    context_path = cwd.joinpath(CONTEXT_FILE_NAME)
    if context_path.exists():
        with context_path.open('r') as f:
            context = json.load(f)
    else:
        raise FileNotFoundError(
            'Could not find \'{}\', are you in a ballet project repo?'
            .format(CONTEXT_FILE_NAME))

    # find out if there are any new keys to prompt for
    with PROJECT_CONTEXT_PATH.open('r') as f:
        new_context = json.load(f)
    new_keys = set(new_context) - set(context['cookiecutter'])
    if new_keys:
        new_context_config = {'cookiecutter': funcy.project(new_context,
                                                            new_keys)}
        new_context = prompt_for_config(new_context_config)
        context['cookiecutter'].update(new_context)

    return context['cookiecutter']


def _call_remote_push(remote):
    return remote.push([
        '{master}:{master}'.format(master=DEFAULT_BRANCH),
        '{project_template}:{project_template}'.format(
            project_template=TEMPLATE_BRANCH),
    ])


@stacklog(logger.info, 'Pushing updates to remote')
def _push(project):
    """Push default branch and project template branch to remote

    With default config (i.e. remote and branch names), equivalent to::

        $ git push origin master:master project-template:project-template

    Raises:
        ballet.exc.BalletError: Push failed in some way
    """
    repo = project.repo
    remote_name = project.config.get('github.remote')
    remote = repo.remote(remote_name)
    result = _call_remote_push(remote)
    failures = lfilter(complement(did_git_push_succeed), result)
    if failures:
        for push_info in failures:
            logger.error(
                'Failed to push ref {from_ref} to {to_ref}'
                .format(from_ref=push_info.local_ref.name,
                        to_ref=push_info.remote_ref.name))
        raise BalletError('Push failed')


def _log_recommended_reinstall():
    logger.info(
        'After a successful project template update, try re-installing the\n'
        'project in case the project template requires any different \n'
        'dependencies than what you have installed:\n'
        '\n'
        '    $ make install')


def update_project_template(push=False, project_template_path=None):
    """Update project with updates to upstream project template

    The update is fairly complicated and proceeds as follows:

    1. Load project: user must run command from master branch and ballet
       must be able to detect the project-template branch
    2. Load the saved cookiecutter context from disk
    3. Render the project template into a temporary directory using the
       saved context, *prompting the user if new keys are required*. Note
       that the project template is simply loaded from the data files of the
       installed version of ballet. Note further that by the project
       template's post_gen_hook, a new git repo is initialized [in the
       temporary directory] and files are committed.
    4. Add the temporary directory as a remote and merge it into the
       project-template branch, favoring changes made to the upstream template.
       Any failure to merge results in an unrecoverable error.
    5. Merge the project-template branch into the master branch. The user is
       responsible for merging conflicts and they are given instructions to
       do so and recover.
    6. If applicable, push to master.

    Args:
        push (bool): whether to push updates to remote, defaults to False
        project_template_path (PathLike): an override for the path to the
            project template
    """
    cwd = pathlib.Path.cwd().resolve()

    # get ballet project info -- must be at project root directory with a
    # ballet.yml file.
    try:
        project = Project.from_path(cwd)
    except ConfigurationError:
        raise ConfigurationError('Must run command from project root.')

    repo = project.repo
    original_head = repo.head.commit.hexsha[:7]

    if repo.is_dirty():
        raise BalletError(
            'Can\'t update project template with uncommitted changes. '
            'Please commit your changes and try again.')

    if repo.head.ref.name != DEFAULT_BRANCH:
        raise ConfigurationError(
            'Must run command from branch {master}'
            .format(master=DEFAULT_BRANCH))

    if TEMPLATE_BRANCH not in repo.branches:
        raise ConfigurationError(
            'Could not find \'{}\' branch.'.format(TEMPLATE_BRANCH))

    # check for upstream updates to ballet
    new_version = _check_for_updated_ballet()
    if new_version:
        _warn_of_updated_ballet(new_version)

    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = pathlib.Path(tempdir)

        # cookiecutter returns path to the resulting project dir
        logger.debug('Re-rendering project template at {}'.format(tempdir))
        updated_template = _render_project_template(
            cwd, tempdir, project_template_path=project_template_path)
        updated_repo = git.Repo(safepath(updated_template))

        # tempdir is a randomly-named dir suitable for a random remote name
        # to avoid conflicts
        remote_name = tempdir.name

        remote = repo.create_remote(
            remote_name, updated_repo.working_tree_dir)
        remote.fetch()

        repo.heads[TEMPLATE_BRANCH].checkout()
        try:
            logger.debug('Merging re-rendered template to project-template '
                         'branch')
            repo.git.merge(
                remote_name + '/' + DEFAULT_BRANCH,
                allow_unrelated_histories=True,
                strategy_option='theirs',
                squash=True,
            )
            if not repo.is_dirty():
                logger.info('No updates to template -- done.')
                return
            commit_message = _make_template_branch_merge_commit_message()
            logger.debug('Committing updates: {}'.format(commit_message))
            repo.git.commit(m=commit_message)
        except GitCommandError:
            logger.critical(
                'Could not merge changes into {template_branch} branch, '
                'update failed'
                .format(template_branch=TEMPLATE_BRANCH))
            raise
        finally:
            _safe_delete_remote(repo, remote_name)
            logger.debug('Checking out master branch')
            repo.heads[DEFAULT_BRANCH].checkout()

    try:
        logger.debug('Merging project-template branch into master')
        repo.git.merge(TEMPLATE_BRANCH, no_ff=True)
    except GitCommandError as e:
        if 'merge conflict' in str(e).lower():
            logger.critical('\n'.join([
                'Update failed due to a merge conflict.',
                'Fix conflicts, and then complete merge manually:',
                '    $ git add .',
                '    $ git commit --no-edit',
                'Otherwise, abandon the update:',
                '    $ git reset --merge {original_head}'
            ]).format(original_head=original_head))
        raise

    if push:
        _push(project)

    _log_recommended_reinstall()



import unittest

import numpy as np
import pandas as pd
from scipy.special import boxcox1p

import ballet.eng.misc
from ballet.util.testing import ArrayLikeEqualityTestingMixin


class TestMisc(ArrayLikeEqualityTestingMixin, unittest.TestCase):
    def setup(self):
        pass

    def test_value_replacer(self):
        trans = ballet.eng.misc.ValueReplacer(0.0, -99)
        data = pd.DataFrame([0, 0, 0, 0, 1, 3, 7, 11, -7])
        expected_result = pd.DataFrame([-99, -99, -99, -99, 1, 3, 7, 11, -7])

        result = trans.fit_transform(data)
        pd.util.testing.assert_frame_equal(result, expected_result)

    def test_box_cox_transformer(self):
        threshold = 0.0
        lmbda = 0.0
        trans = ballet.eng.misc.BoxCoxTransformer(threshold=threshold,
                                                  lmbda=lmbda)

        skewed = [0., 0., 0., 0., 1.]
        unskewed = [0., 0., 0., 0., 0.]

        exp_skew_res = boxcox1p(skewed, lmbda)
        exp_unskew_res = unskewed

        # test on DF, one skewed column
        df = pd.DataFrame()
        df['skewed'] = skewed
        df['unskewed'] = unskewed
        df_res = trans.fit_transform(df)
        self.assertIsInstance(df_res, pd.DataFrame)
        self.assertIn('skewed', df_res.columns)
        self.assertIn('unskewed', df_res.columns)
        self.assertArrayAlmostEqual(df_res['skewed'], exp_skew_res)

        # test on DF, no skewed columns
        df_unskewed = pd.DataFrame()
        df_unskewed['col1'] = unskewed
        df_unskewed['col2'] = unskewed
        df_unskewed_res = trans.fit_transform(df_unskewed)
        self.assertIsInstance(df_unskewed_res, pd.DataFrame)
        self.assertIn('col1', df_unskewed_res.columns)
        self.assertIn('col2', df_unskewed_res.columns)
        self.assertArrayEqual(df_unskewed_res['col1'], exp_unskew_res)
        self.assertArrayEqual(df_unskewed_res['col2'], exp_unskew_res)

        # test on skewed Series
        ser_skewed = pd.Series(skewed)
        ser_skewed_res = trans.fit_transform(ser_skewed)
        self.assertIsInstance(ser_skewed_res, pd.Series)
        self.assertArrayAlmostEqual(ser_skewed_res, exp_skew_res)

        # test on unskewed Series
        ser_unskewed = pd.Series(unskewed)
        ser_unskewed_res = trans.fit_transform(ser_unskewed)
        self.assertIsInstance(ser_unskewed_res, pd.Series)
        self.assertArrayEqual(ser_unskewed_res, exp_unskew_res)

        # test on array
        arr = np.array([unskewed, skewed]).T
        arr_res = trans.fit_transform(arr)
        arr_exp = np.vstack((exp_unskew_res, exp_skew_res)).T
        self.assertArrayAlmostEqual(arr_res, arr_exp)

    def test_named_framer(self):
        name = 'foo'

        # good objects
        index = pd.Index([1, 2, 3])
        ser = pd.Series([1, 2, 3])
        df = pd.DataFrame([1, 2, 3])
        arr = np.array([1, 2, 3])

        for obj in [index, ser, df, arr]:
            trans = ballet.eng.misc.NamedFramer(name)
            result = trans.fit_transform(obj)
            self.assertIsInstance(result, pd.DataFrame)
            self.assertEqual(result.shape[1], 1)
            self.assertEqual(result.columns, [name])

        # bad objects -- wrong type
        int_ = 1
        str_ = 'hello'
        for obj in [int_, str_]:
            trans = ballet.eng.misc.NamedFramer(name)
            with self.assertRaises(TypeError):
                result = trans.fit_transform(obj)

        # bad objects -- too wide
        wide_df = pd.DataFrame(np.arange(10).reshape(-1, 2))
        wide_arr = np.arange(10).reshape(-1, 2)
        for obj in [wide_df, wide_arr]:
            trans = ballet.eng.misc.NamedFramer(name)
            with self.assertRaises(ValueError):
                result = trans.fit_transform(obj)



import pathlib
from unittest.mock import MagicMock, create_autospec, patch

import numpy as np
import pandas as pd
from funcy import contextmanager

from ballet.project import Project
from ballet.util.git import make_commit_range
from ballet.validation.common import ChangeCollector
from ballet.validation.feature_api.validator import FeatureApiValidator
from ballet.validation.project_structure.validator import (
    ProjectStructureValidator)

from ..util import make_mock_commit, mock_repo


class SampleDataMixin:
    def setUp(self):
        self.df = pd.DataFrame(
            data={
                'country': ['USA', 'USA', 'Canada', 'Japan'],
                'year': [2001, 2002, 2001, 2002],
                'size': [np.nan, -11, 12, 0.0],
                'strength': [18, 110, np.nan, 101],
                'happy': [False, True, False, False]
            }
        ).set_index(['country', 'year'])
        self.X = self.df[['size', 'strength']]
        self.y = self.df[['happy']]
        super().setUp()


def make_mock_project(repo, pr_num, path, contrib_module_path):
    def mock_get(key):
        if key == 'contrib.module_path':
            return contrib_module_path
        else:
            raise KeyError
    config = MagicMock()
    config.get.side_effect = mock_get

    project = create_autospec(Project)
    project.repo = repo
    project.pr_num = str(pr_num)
    project.path = pathlib.Path(path)
    project.config = config

    # in attr map
    project.load_data = MagicMock()
    project.build = MagicMock()
    project.collect_contrib_features = MagicMock()

    return project


@contextmanager
def mock_project_content(path_content):
    with mock_repo() as repo:
        for path, content in path_content:
            make_mock_commit(repo, path=path, content=content)
        yield repo


@contextmanager
def null_change_collector(pr_num):
    with mock_repo() as repo:
        commit_range = 'HEAD^..HEAD'
        contrib_module_path = None

        travis_env_vars = {
            'TRAVIS_BUILD_DIR': repo.working_tree_dir,
            'TRAVIS_PULL_REQUEST': str(pr_num),
            'TRAVIS_COMMIT_RANGE': commit_range,
        }

        with patch.dict('os.environ', travis_env_vars, clear=True):
            project_path = repo.working_tree_dir
            project = make_mock_project(repo, pr_num, project_path,
                                        contrib_module_path)
            yield ChangeCollector(project)


@contextmanager
def mock_file_change_validator(
    path_content, pr_num, contrib_module_path
):
    """ProjectStructureValidator for mock repo and mock project content

    Args:
        path_content: iterable of (relative path, file content)
    """
    with mock_project_content(path_content) as repo:
        travis_build_dir = repo.working_tree_dir
        travis_pull_request = str(pr_num)
        travis_commit_range = '{}...{}'.format(
            repo.head.commit.parents[0], repo.head.commit)

        travis_env_vars = {
            'TRAVIS_BUILD_DIR': travis_build_dir,
            'TRAVIS_PULL_REQUEST': travis_pull_request,
            'TRAVIS_COMMIT_RANGE': travis_commit_range,
        }

        with patch.dict('os.environ', travis_env_vars, clear=True):
            project_path = repo.working_tree_dir
            project = make_mock_project(repo, pr_num, project_path,
                                        contrib_module_path)
            yield ProjectStructureValidator(project)


@contextmanager
def mock_feature_api_validator(
    path_content, pr_num, contrib_module_path, X, y
):
    """ProjectStructureValidator for mock repo and mock project content

    Args:
        path_content: iterable of (relative path, file content)
    """
    with mock_project_content(path_content) as repo:
        travis_build_dir = repo.working_tree_dir
        travis_pull_request = str(pr_num)
        travis_commit_range = make_commit_range(
            repo.head.commit.parents[0], repo.head.commit)

        travis_env_vars = {
            'TRAVIS_BUILD_DIR': travis_build_dir,
            'TRAVIS_PULL_REQUEST': travis_pull_request,
            'TRAVIS_COMMIT_RANGE': travis_commit_range,
        }

        with patch.dict('os.environ', travis_env_vars, clear=True):
            project_path = repo.working_tree_dir
            project = make_mock_project(repo, pr_num, project_path,
                                        contrib_module_path)
            project.load_data.return_value = X, y
            yield FeatureApiValidator(project)



import click
from stacklog import stacklog

from ballet import __version__ as version


@click.group()
@click.version_option(version)
@click.option('-v', '--verbose',
              count=True,
              help='Increase verbosity')
@click.option('-q', '--quiet',
              count=True,
              help='Decrease verbosity'
              )
def cli(verbose, quiet):
    """The Ballet command line interface"""
    import ballet.util.log
    # Process logging
    count = verbose - quiet
    if count <= -1:
        level = 'CRITICAL'
    elif count == 0:
        level = 'INFO'
    else:
        level = 'DEBUG'
    ballet.util.log.enable(level=level,
                           format=ballet.util.log.SIMPLE_LOG_FORMAT,
                           echo=False)


@cli.command()
@stacklog(click.echo, 'Generating new ballet project')
def quickstart():
    """Generate a brand-new ballet project"""
    import ballet.templating
    ballet.templating.render_project_template()


@cli.command('update-project-template')
@click.option('--push/--no-push', '-p',
              default=False,
              help='Push updates to remote on success')
@click.option('--path', 'project_template_path',
              default=None,
              help='Specify override for project template path '
                   '(i.e. gh:HDI-Project/ballet)')
@stacklog(click.echo, 'Updating project template')
def update_project_template(push, project_template_path):
    """Update an existing ballet project from the upstream template"""
    import ballet.update
    ballet.update.update_project_template(
        push=push, project_template_path=project_template_path)


@cli.command('start-new-feature')
@stacklog(click.echo, 'Starting new feature')
def start_new_feature():
    """Start working on a new feature from a template"""
    import ballet.templating
    ballet.templating.start_new_feature()


@cli.command('validate')
@click.option('--check-all', '-A',
              is_flag=True,
              default=False)
@click.option('--check-project-structure/--no-check-project-structure',
              envvar='BALLET_CHECK_PROJECT_STRUCTURE',
              default=False)
@click.option('--check-feature-api/--no-check-feature-api',
              envvar='BALLET_CHECK_FEATURE_API',
              default=False)
@click.option('--evaluate-feature-acceptance/--no-evaluate-feature-acceptance',
              envvar='BALLET_EVALUATE_FEATURE_ACCEPTANCE',
              default=False)
@click.option('--evaluate-feature-pruning/--no-evaluate-feature-pruning',
              envvar='BALLET_EVALUATE_FEATURE_PRUNING',
              default=False)
@stacklog(click.echo, 'Validating project')
def validate(check_all, check_project_structure, check_feature_api,
             evaluate_feature_acceptance, evaluate_feature_pruning):
    """Validate project changes from a candidate branch"""
    import pathlib

    import ballet.util.log
    import ballet.project
    import ballet.validation.main

    # over-write logging settings?
    ballet.util.log.enable(level='DEBUG',
                           format=ballet.util.log.SIMPLE_LOG_FORMAT,
                           echo=False)

    cwd = pathlib.Path.cwd()
    project = ballet.project.Project.from_path(cwd)
    ballet.validation.main.validate(
        project,
        check_project_structure or check_all,
        check_feature_api or check_all,
        evaluate_feature_acceptance or check_all,
        evaluate_feature_pruning or check_all)



import unittest
from unittest.mock import Mock, patch

import git

import ballet
import ballet.util.log
from ballet.update import (
    _check_for_updated_ballet, _extract_latest_from_search_triple,
    _get_latest_ballet_version_string, _log_recommended_reinstall,
    _make_template_branch_merge_commit_message, _query_pip_search_ballet,
    _safe_delete_remote, _warn_of_updated_ballet)


class UpdateTest(unittest.TestCase):

    def test_query_pip_search_ballet(self):
        # nothing better to do that just call the function...
        # actually hits PyPI but difficult to mock because uses subprocess >:(
        result = _query_pip_search_ballet()
        self.assertIn('ballet', result)

    def test_extract_latest_from_search_triple(self):
        triple = (
            'ballet (x.y.z)  - some description',
            '  INSTALLED: x.y.z',
            '  LATEST:    u.v.w',
        )
        result = _extract_latest_from_search_triple(triple)
        self.assertEqual(result, 'u.v.w')

    def test_extract_latest_from_search_triple_not_found(self):
        triple = (
            'something',
            'that does not',
            'match the expected format'
        )
        result = _extract_latest_from_search_triple(triple)
        self.assertIsNone(result)

    @patch('funcy.partition')
    @patch('ballet.update._extract_latest_from_search_triple')
    @patch('ballet.update._query_pip_search_ballet')
    def test_get_latest_ballet_version_string(self, mock_query,
                                              mock_extract, mock_partition):
        mock_partition.return_value = ['some', 'iterable']
        expected = 'x.y.z'
        mock_extract.return_value = expected

        actual = _get_latest_ballet_version_string()

        self.assertEqual(actual, expected)

    @patch('ballet.update._get_latest_ballet_version_string')
    def test_check_for_updated_ballet(self, mock_latest):
        # obviously this will represent an update from whatever the current
        # version is
        latest = '99999999999.0.0'
        mock_latest.return_value = latest
        expected = latest
        actual = _check_for_updated_ballet()
        self.assertEqual(actual, expected)

    @patch('ballet.update._get_latest_ballet_version_string')
    def test_check_for_updated_ballet_no_updates(self, mock_latest):
        mock_latest.return_value = ballet.__version__
        expected = None  # no updates available
        actual = _check_for_updated_ballet()
        self.assertEqual(actual, expected)

    def test_warn_of_updated_ballet(self):
        latest = 'x.y.z'
        logger = ballet.util.log.logger
        with self.assertLogs(logger) as cm:
            _warn_of_updated_ballet(latest)
        output = '\n'.join(cm.output)
        self.assertIn(latest, output)

    def test_make_template_branch_merge_commit_message(self):
        result = _make_template_branch_merge_commit_message()
        self.assertIn(ballet.__version__, result)

    def test_safe_delete_remote(self):
        repo = Mock(spec=git.Repo)
        name = 'name'
        _safe_delete_remote(repo, name)
        repo.delete_remote.assert_called_once_with(name)

    def test_log_recommended_reinstall(self):
        logger = ballet.util.log.logger
        with self.assertLogs(logger=logger):
            _log_recommended_reinstall()



import funcy
import numpy as np
import pandas as pd
import sklearn.base
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from ballet.exc import BalletError
from ballet.util import get_arr_desc

__all__ = [
    'BaseTransformer',
    'ConditionalTransformer',
    'GroupedFunctionTransformer',
    'GroupwiseTransformer',
    'NoFitMixin',
    'SimpleFunctionTransformer',
]


class NoFitMixin:
    """Mix-in class for transformations that do not require a fit stage"""

    def fit(self, X, y=None, **fit_kwargs):
        return self


class BaseTransformer(NoFitMixin, TransformerMixin, BaseEstimator):
    """Base transformer class for developing new transformers"""
    pass


class SimpleFunctionTransformer(BaseTransformer):
    """Transformer that applies a callable to its input

    The callable will be called on the input X in the transform stage,
    optionally with additional arguments and keyword arguments.

    This will be eventually replaced with
    ``sklearn.preprocessing.FunctionTransformer``.

    Args:
        func (callable): callable to apply
        func_args (tuple): additional arguments to pass
        func_kwargs (dict): keyword arguments to pass
    """

    def __init__(self, func, func_args=None, func_kwargs=None):
        super().__init__()
        self.func = func
        self.func_args = func_args
        self.func_kwargs = func_kwargs

    def _func_call(self, *args):
        func_args = self.func_args or ()
        func_kwargs = self.func_kwargs or {}
        args = funcy.join((args, func_args))  # py34
        return self.func(*args, **func_kwargs)

    def transform(self, X, **transform_kwargs):
        return self._func_call(X)


class GroupedFunctionTransformer(SimpleFunctionTransformer):
    """Transformer that applies a callable to each group of a groupby

    Args:
        func (callable): callable to apply
        func_args (tuple): additional arguments to pass
        func_kwargs (dict): keyword arguments to pass
        groupby_kwargs (dict): keyword arguments to pd.DataFrame.groupby
    """

    def __init__(self, func, func_args=None,
                 func_kwargs=None, groupby_kwargs=None):
        super().__init__(func, func_args=func_args, func_kwargs=func_kwargs)
        self.groupby_kwargs = groupby_kwargs

    def transform(self, X, **transform_kwargs):
        groupby_kwargs = self.groupby_kwargs or {}
        if groupby_kwargs:
            call = X.groupby(**groupby_kwargs).apply
        else:
            call = X.pipe
        return call(self._func_call)


class GroupwiseTransformer(BaseTransformer):
    """Transformer that does something different for every group

    For each group identified in the training set by the groupby operation,
    a separate transformer is cloned and fit. This is useful to learn
    group-wise transformers that do not leak data between the training and
    test sets. Consider the case of imputing missing values with the mean of
    some group. A normal, pure-pandas implementation, such as
    ``X_te.groupby(by='foo').apply('mean')`` would leak information about
    the test set means, which might differ from the training set means.

    Args:
        transformer (transformer-like or callable): the transformer to apply
            to each group. If transformer is a transformer-like instance (i.e.
            has fit, transform methods etc.), then it is cloned for each group.
            If transformer is a transformer-like class (i.e. instances of
            the class are transformer-like), then it is initialized with no
            arguments for each group. If it is a callable, then it is called
            with no arguments for each group.
        groupby_kwargs (dict): keyword arguments to pd.DataFrame.groupby
        column_selection (str, List[str]): column, or list of columns,
            to select after the groupby. Equivalent to
            `df.groupby(...)[column_selection]`. Defaults to None, i.e. no
            column selection is performed.
        handle_unknown (str): 'error' or 'ignore', default=’error’. Whether to
            raise an error or ignore if an unknown group is encountered during
            transform. When this parameter is set to 'ignore' and an unknown
            group is encountered during transform, the group's values will be
            passed through unchanged.
        handle_error (str): 'error' or 'ignore', default='error'. Whether to
            raise an error or ignore if an error is raised during transforming
            an individual group. When this parameter is set to 'ignore' and
            an error is raised when calling the transformer's transform
            method on an individual group, the group's values will be passed
            through unchanged.

    Example usage:

        In this example, we create a groupwise transformer that fits a
        separate imputer for each group encountered. For new data points,
        values will be imputed according to the mean of its group on the
        training set, avoiding any data leakage.

        .. code-block:: python

           >>> from sklearn.impute import SimpleImputer
           >>> transformer = GroupwiseTransformer(
           ...     SimpleImputer(strategy='mean'),
           ...     groupby_kwargs = {'level': 'name'}
           ... )

    Raises:
        ballet.exc.BalletError: if handle_unknown=='error' and an unknown group
            is encountered at transform-time.
    """

    def __init__(self,
                 transformer,
                 groupby_kwargs=None,
                 column_selection=None,
                 handle_unknown='error',
                 handle_error='error'):
        self.transformer = transformer
        self.groupby_kwargs = groupby_kwargs
        self.column_selection = column_selection
        self.handle_unknown = handle_unknown
        self.handle_error = handle_error

    def _make_transformer(self):
        if isinstance(self.transformer, type) or callable(self.transformer):
            return self.transformer()
        else:
            return sklearn.base.clone(self.transformer)

    def fit(self, X, y=None, **fit_kwargs):
        # validation on inputs
        self.groupby_kwargs_ = self.groupby_kwargs or {}
        if self.handle_unknown not in ['error', 'ignore']:
            raise ValueError(
                'Invalid value for handle_unknown: {}'
                .format(self.handle_unknown))
        if self.handle_error not in ['error', 'ignore']:
            raise ValueError(
                'Invalid value for handle_error: {}'
                .format(self.handle_error))

        # Get the groups
        grouper = X.groupby(**self.groupby_kwargs_)
        self.groups_ = set(grouper.groups.keys())

        # Create and fit a transformer for each group
        self.transformers_ = {}
        for group_name, x_group in grouper:
            transformer = self._make_transformer()

            if self.column_selection is not None:
                x_group = x_group[self.column_selection]

            if y is not None:
                # Extract y by integer indexing
                inds = grouper.indices[group_name]
                y_group = y[inds]
                transformer.fit(x_group, y_group)
            else:
                transformer.fit(x_group)

            self.transformers_[group_name] = transformer

        return self

    def transform(self, X, **transform_kwargs):
        check_is_fitted(self, ['groups_', 'transformers_'])

        def _transform(x_group, *args, **kwargs):
            # If the group is not a DataFrame, there are two problems
            # 1. We can't rely on group.name to lookup the right transformer
            # 2. We can't "reassemble" the transformed
            # However, the contract of ``pandas.core.groupby.GroupBy.apply`` is
            # that the input is a DataFrame, so this should never occur.
            if not isinstance(x_group, pd.DataFrame):
                raise NotImplementedError

            group_name = x_group.name

            if self.column_selection is not None:
                x_group = x_group[self.column_selection]

            if group_name in self.transformers_:
                transformer = self.transformers_[group_name]
                try:
                    data = transformer.transform(x_group, *args, **kwargs)

                    # This post-processing step is required because sklearn
                    # transform converts a DataFrame to an array. This is my
                    # best attempt so far to approximate the following:
                    # >>> result = x_group.copy()
                    # >>> result.values = data
                    # which is an error as `values` cannot be set.
                    index = x_group.index
                    columns = x_group.columns
                    return pd.DataFrame(
                        data=data, index=index, columns=columns)
                except Exception:
                    if self.handle_error == 'ignore':
                        return x_group
                    else:
                        raise
            else:
                if self.handle_unknown == 'error':
                    raise BalletError(
                        'Unknown group: {group_name}'
                        .format(group_name=group_name))
                elif self.handle_unknown == 'ignore':
                    return x_group
                else:
                    # Unreachable code
                    raise RuntimeError

        return (
            X
            .groupby(**self.groupby_kwargs_)
            .apply(_transform, **transform_kwargs)
        )


class ConditionalTransformer(BaseTransformer):
    """Transform columns that satisfy a condition during training

    In the fit stage, determines which variables (columns) satisfy the
    condition. In the transform stage, applies the given transformation to
    the selected columns, passing through the complement unchanged.

    Args:
        condition (callable): condition function
        trans (callable): transform function
    """

    def __init__(self, condition, trans):
        super().__init__()
        self.condition = condition
        # 1. can't call it transform because conflicts with method name
        # 2. can't call it transform and assign to a different attribute
        # because sklearn.clone would complain
        self.trans = trans

    def fit(self, X, y=None, **fit_args):
        # features_to_transform_ is a bool or array[bool]
        self.features_to_transform_ = self.condition(X)
        return self

    def transform(self, X, **transform_args):
        check_is_fitted(self, 'features_to_transform_')

        if isinstance(X, pd.DataFrame):
            X = X.copy()
            X.loc[:, self.features_to_transform_] = self.trans(
                X.loc[:, self.features_to_transform_])
            return X
        elif np.ndim(X) == 1:
            return self.trans(X) if self.features_to_transform_ else X
        elif isinstance(X, np.ndarray):
            if self.features_to_transform_.any():
                X = X.copy()
                X = X.astype('float')
                mask = np.tile(self.features_to_transform_, (X.shape[0], 1))
                np.putmask(X, mask, self.trans(
                    X[:, self.features_to_transform_]))
            return X
        elif not self.features_to_transform_:
            # if we wouldn't otherwise have known what to do, we can pass
            # through X if transformation was not necessary anyways
            return X
        else:
            raise TypeError(
                "Couldn't apply transformer on features in {}."
                .format(get_arr_desc(X)))



import unittest

import funcy
import numpy as np
import pandas as pd
import sklearn.preprocessing

from ballet.compat import SimpleImputer
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.pipeline import FeatureEngineeringPipeline
from ballet.transformer import DelegatingRobustTransformer
from ballet.util import asarray2d

from .util import FragileTransformer, FragileTransformerPipeline


class DelegatingRobustTransformerTest(unittest.TestCase):

    def setUp(self):
        self.X_ser = pd.util.testing.makeFloatSeries()
        self.X_df = self.X_ser.to_frame()
        self.X_arr1d = np.asarray(self.X_ser)
        self.X_arr2d = np.asarray(self.X_df)
        self.y_ser = self.X_ser.copy()
        self.y_df = self.X_df.copy()
        self.y_arr1d = np.asarray(self.y_ser)
        self.y_arr2d = np.asarray(self.y_df)

        self.d = {
            'ser': (self.X_ser, self.y_ser),
            'df': (self.X_df, self.y_df),
            'arr1d': (self.X_arr1d, self.y_arr1d),
            'arr2d': (self.X_arr2d, self.y_arr2d),
        }

    def _test_robust_transformer(
        self, input_types, bad_input_checks, catches,
        transformer_maker=FragileTransformer
    ):
        fragile_transformer = transformer_maker(bad_input_checks, catches)
        robust_transformer = DelegatingRobustTransformer(
            transformer_maker(bad_input_checks, catches))

        for input_type in input_types:
            X, y = self.d[input_type]
            # fragile transformer raises error
            with self.assertRaises(catches):
                fragile_transformer.fit_transform(X, y)
            # robust transformer does not raise error
            X_robust = robust_transformer.fit_transform(X, y)
            self.assertTrue(
                np.array_equal(
                    asarray2d(X),
                    asarray2d(X_robust)
                )
            )

    def test_robust_transformer_ser(self):
        input_types = ('ser',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer(input_types, bad_input_checks, catches)

    def test_robust_transformer_df(self):
        input_types = ('ser', 'df',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer(input_types, bad_input_checks, catches)

    def test_robust_transformer_arr(self):
        input_types = ('ser', 'df', 'arr1d')
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
            lambda x: isinstance(x, np.ndarray) and x.ndim == 1,
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer(input_types, bad_input_checks, catches)

    def test_robust_transformer_sklearn(self):
        Transformers = (
            SimpleImputer,
            sklearn.preprocessing.StandardScaler,
            sklearn.preprocessing.Binarizer,
            sklearn.preprocessing.PolynomialFeatures,
        )
        # some of these input types are bad for sklearn.
        input_types = ('ser', 'df', 'arr1d')
        for Transformer in Transformers:
            robust_transformer = DelegatingRobustTransformer(Transformer())
            for input_type in input_types:
                X, y = self.d[input_type]
                robust_transformer.fit_transform(X, y=y)

    def _test_robust_transformer_pipeline(
        self, input_types, bad_input_checks, catches
    ):
        FragileTransformerPipeline3 = funcy.partial(
            FragileTransformerPipeline, 3)
        return self._test_robust_transformer(
            input_types,
            bad_input_checks,
            catches,
            transformer_maker=FragileTransformerPipeline3)

    def test_robust_transformer_pipeline_ser(self):
        input_types = ('ser',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer_pipeline(
            input_types, bad_input_checks, catches)

    def test_robust_transformer_pipeline_df(self):
        input_types = ('ser', 'df',)
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer_pipeline(
            input_types, bad_input_checks, catches)

    def test_robust_transformer_pipeline_arr(self):
        input_types = ('ser', 'df', 'arr1d')
        bad_input_checks = (
            lambda x: isinstance(x, pd.Series),
            lambda x: isinstance(x, pd.DataFrame),
            lambda x: isinstance(x, np.ndarray) and x.ndim == 1,
        )
        catches = (ValueError, TypeError)
        self._test_robust_transformer_pipeline(
            input_types, bad_input_checks, catches)


class FeatureTest(unittest.TestCase):

    def setUp(self):
        self.input = 'foo'
        self.transformer = IdentityTransformer()

    def test_feature_init_with_one_transformer(self):
        Feature(self.input, self.transformer)

    def test_feature_init_with_list_of_transformer(self):
        Feature(self.input, [self.transformer, self.transformer])

    def test_feature_init_with_callable(self):
        Feature(self.input, funcy.identity)

    def test_feature_init_with_list_of_transformers_and_callables(self):
        Feature(self.input, [self.transformer, funcy.identity])

    def test_feature_init_invalid_transformer_api(self):
        with self.assertRaises(ValueError):
            Feature(self.input, object())

        with self.assertRaises(ValueError):
            Feature(self.input, IdentityTransformer)

    def test_feature_as_input_transformer_tuple(self):
        feature = Feature(self.input, self.transformer)
        tup = feature.as_input_transformer_tuple()
        self.assertIsInstance(tup, tuple)
        self.assertEqual(len(tup), 3)

    def test_feature_as_feature_engineering_pipeline(self):
        feature = Feature(self.input, self.transformer)
        mapper = feature.as_feature_engineering_pipeline()
        self.assertIsInstance(mapper, FeatureEngineeringPipeline)


class FeatureEngineeringPipelineTest(FeatureTest):

    def test_init_seqcont(self):
        feature = Feature(self.input, self.transformer)
        features = [feature]
        mapper = FeatureEngineeringPipeline(features)
        self.assertIsInstance(mapper, FeatureEngineeringPipeline)

    def test_init_scalar(self):
        feature = Feature(self.input, self.transformer)
        mapper = FeatureEngineeringPipeline(feature)
        self.assertIsInstance(mapper, FeatureEngineeringPipeline)

    def test_fit(self):
        feature = Feature(self.input, self.transformer)
        mapper = FeatureEngineeringPipeline(feature)
        df = pd.util.testing.makeCustomDataframe(5, 2)
        df.columns = ['foo', 'bar']
        mapper.fit(df)

    def test_transform(self):
        feature = Feature(self.input, self.transformer)
        mapper = FeatureEngineeringPipeline(feature)
        df = pd.util.testing.makeCustomDataframe(5, 2)
        df.columns = ['foo', 'bar']
        mapper.fit(df)
        X = mapper.transform(df)
        self.assertEqual(np.shape(X), (5, 1))



from ballet.contrib import _collect_contrib_features
from ballet.util.log import logger
from ballet.validation.base import BaseValidator
from ballet.validation.common import (
    ChangeCollector, subsample_data_for_validation)
from ballet.validation.feature_api import validate_feature_api


class FeatureApiValidator(BaseValidator):

    def __init__(self, project):
        self.change_collector = ChangeCollector(project)

        X, y = project.load_data()
        self.X, self.y = subsample_data_for_validation(X, y)

    def validate(self):
        """Collect and validate all new features"""

        changes = self.change_collector.collect_changes()

        features = []
        imported_okay = True
        for importer, modname, modpath in changes.new_feature_info:
            try:
                mod = importer()
                features.extend(_collect_contrib_features(mod))
            except (ImportError, SyntaxError):
                logger.info(
                    'Failed to import module at {}'
                    .format(modpath))
                logger.exception('Exception details: ')
                imported_okay = False

        if not imported_okay:
            return False

        # if no features were added at all, reject
        if not features:
            logger.info('Failed to collect any new features.')
            return False

        return all(
            validate_feature_api(feature, self.X, self.y, subsample=False)
            for feature in features
        )



import pathlib
from collections import namedtuple

from funcy import collecting, complement, lfilter, partial, post_processing
from stacklog import stacklog

from ballet.contrib import _collect_contrib_feature_from_module
from ballet.exc import (
    BalletError, FeatureCollectionError, NoFeaturesCollectedError)
from ballet.util import make_plural_suffix, whether_failures
from ballet.util.ci import TravisPullRequestBuildDiffer, can_use_travis_differ
from ballet.util.git import LocalMergeBuildDiffer, LocalPullRequestBuildDiffer
from ballet.util.log import logger
from ballet.util.mod import import_module_at_path, relpath_to_modname
from ballet.validation.base import FeaturePerformanceEvaluator
from ballet.validation.project_structure.checks import ProjectStructureCheck


def get_proposed_feature(project):
    """Get the proposed feature

    The path of the proposed feature is determined by diffing the project
    against a comparison branch, such as master. The feature is then imported
    from that path and returned.

    Args:
        project (ballet.project.Project): project info

    Raises:
        ballet.exc.BalletError: more than one feature collected
    """
    change_collector = ChangeCollector(project)
    collected_changes = change_collector.collect_changes()
    new_feature_info = collected_changes.new_feature_info

    n = len(new_feature_info)
    if n == 0:
        raise NoFeaturesCollectedError
    elif n == 1:
        importer, _, _ = new_feature_info[0]
        module = importer()
        feature = _collect_contrib_feature_from_module(module)
        return feature
    else:
        msg = 'Too many features collected (n={})'.format(n)
        raise FeatureCollectionError(msg)


def get_accepted_features(features, proposed_feature):
    """Deselect candidate features from list of all features

    Args:
        features (List[Feature]): collection of all features in the ballet
            project: both accepted features and candidate ones that have not
            been accepted
        proposed_feature (Feature): candidate feature that has not been
            accepted

    Returns:
        List[Feature]: list of features with the proposed feature not in it.

    Raises:
        ballet.exc.BalletError: Could not deselect exactly the proposed
            feature.
    """
    def eq(feature):
        """Features are equal if they have the same source

        At least in this implementation...
        """
        return feature.source == proposed_feature.source

    # deselect features that match the proposed feature
    result = lfilter(complement(eq), features)

    if len(features) - len(result) == 1:
        return result
    elif len(result) == len(features):
        raise BalletError(
            'Did not find match for proposed feature within \'contrib\'')
    else:
        raise BalletError(
            'Unexpected condition (n_features={}, n_result={})'
            .format(len(features), len(result)))


def _log_collect_items(name, items):
    n = len(items)
    s = make_plural_suffix(items)
    logger.info('Collected {n} {name}{s}'.format(n=n, name=name, s=s))
    return items


CollectedChanges = namedtuple(
    'CollectedChanges',
    'file_diffs candidate_feature_diffs valid_init_diffs inadmissible_diffs '
    'new_feature_info')


class ChangeCollector:
    """Validate the features introduced in a proposed pull request.

    Args:
        project (ballet.project.Project): project info
        differ (ballet.util.git.Differ, optional): specific differ to use;
            if not provided, will be determined automatically from the
            environment
    """

    def __init__(self, project, differ=None):
        self.project = project
        self.differ = differ

        if self.differ is None:
            pr_num = self.project.pr_num
            repo = self.project.repo
            if pr_num is None:
                self.differ = LocalMergeBuildDiffer(repo)
            elif can_use_travis_differ():
                self.differ = TravisPullRequestBuildDiffer(pr_num)
            else:
                self.differ = LocalPullRequestBuildDiffer(pr_num, repo)

    def collect_changes(self):
        """Collect file and feature changes

        Steps

        1. Collects the files that have changed in this pull request as
           compared to a comparison branch.
        2. Categorize these file changes into admissible or inadmissible file
           changes. Admissible file changes solely contribute python files to
           the contrib subdirectory.
        3. Collect features from admissible new files.

        Returns:
            CollectedChanges
        """

        file_diffs = self._collect_file_diffs()
        candidate_feature_diffs, valid_init_diffs, inadmissible_diffs = \
            self._categorize_file_diffs(file_diffs)
        new_feature_info = self._collect_feature_info(candidate_feature_diffs)

        return CollectedChanges(
            file_diffs, candidate_feature_diffs, valid_init_diffs,
            inadmissible_diffs, new_feature_info)

    @post_processing(partial(_log_collect_items, 'file'))
    @stacklog(logger.info, 'Collecting file changes')
    def _collect_file_diffs(self):
        file_diffs = self.differ.diff()

        # log results
        for i, file in enumerate(file_diffs):
            logger.debug('File {i}: {file}'.format(i=i, file=file))

        return file_diffs

    @stacklog(logger.info, 'Categorizing file changes')
    def _categorize_file_diffs(self, file_diffs):
        """Partition file changes into admissible and inadmissible changes"""
        # TODO move this into a new validator
        candidate_feature_diffs = []
        valid_init_diffs = []
        inadmissible_files = []

        for diff in file_diffs:
            valid, failures = check_from_class(
                ProjectStructureCheck, diff, self.project)
            if valid:
                if pathlib.Path(diff.b_path).parts[-1] != '__init__.py':
                    candidate_feature_diffs.append(diff)
                    logger.debug(
                        'Categorized {file} as CANDIDATE FEATURE MODULE'
                        .format(file=diff.b_path))
                else:
                    valid_init_diffs.append(diff)
                    logger.debug(
                        'Categorized {file} as VALID INIT MODULE'
                        .format(file=diff.b_path))
            else:
                inadmissible_files.append(diff)
                logger.debug(
                    'Categorized {file} as INADMISSIBLE; '
                    'failures were {failures}'
                    .format(file=diff.b_path, failures=failures))

        logger.info(
            'Admitted {} candidate feature{} '
            'and {} __init__ module{} '
            'and rejected {} file{}'
            .format(len(candidate_feature_diffs),
                    make_plural_suffix(candidate_feature_diffs),
                    len(valid_init_diffs),
                    make_plural_suffix(valid_init_diffs),
                    len(inadmissible_files),
                    make_plural_suffix(inadmissible_files)))

        return candidate_feature_diffs, valid_init_diffs, inadmissible_files

    @post_processing(partial(_log_collect_items, 'feature'))
    @collecting
    @stacklog(logger.info, 'Collecting info on newly-proposed features')
    def _collect_feature_info(self, candidate_feature_diffs):
        """Collect feature info

        Args:
            candidate_feature_diffs (List[git.diff.Diff]): list of Diffs
                corresponding to admissible file changes compared to
                comparison ref

        Returns:
            List[Tuple[Callable, str, str]]: list of tuple of importer,
                module name, and module path. The "importer" is a callable that
                returns a module
        """

        # the directory containing ballet.yml
        project_root = self.project.path

        # the directory containing the package
        try:
            package_path = self.project.package.__path__[0]
            package_root = pathlib.Path(package_path).parent
        except (AttributeError, IndexError):
            logger.debug("Couldn't get package root, will try to recover",
                         exc_info=True)
            package_root = project_root

        for diff in candidate_feature_diffs:
            path = diff.b_path
            relpath = project_root.joinpath(path).relative_to(package_root)
            modname = relpath_to_modname(relpath)
            modpath = project_root.joinpath(path)
            importer = partial(import_module_at_path, modname, modpath)
            yield importer, modname, modpath


def subsample_data_for_validation(X, y):
    return X, y


@whether_failures
def check_from_class(check_class, obj, *checker_args, **checker_kwargs):
    for Checker in check_class.__subclasses__():
        check = Checker(*checker_args, **checker_kwargs).do_check
        success = check(obj)
        if not success:
            name = Checker.__name__
            yield name


class RandomFeaturePerformanceEvaluator(FeaturePerformanceEvaluator):

    def __init__(self, *args, p=0.3, seed=None):
        super().__init__(*args)
        self.p = p
        self.seed = seed

    def __str__(self):
        return '{str}: p={p}, seed={seed}'.format(
            str=super().__str__(), p=self.p, seed=self.seed)



import pandas as pd

from ballet.eng.base import BaseTransformer, GroupedFunctionTransformer

__all__ = ['LagImputer', 'NullFiller', 'NullIndicator']


class LagImputer(GroupedFunctionTransformer):
    """Fill missing values using group-specific lags"""

    def __init__(self, groupby_kwargs=None):
        super().__init__(lambda x: x.fillna(method='ffill'),
                         groupby_kwargs=groupby_kwargs)


class NullFiller(BaseTransformer):
    """Fill values passing a filter with a given replacement

    Args:
        isnull (callable): vectorized test of whether a value is consider null.
            Defaults to ``pandas.isnull``.
        replacement: replacement for each null value
    """

    def __init__(self, isnull=pd.isnull, replacement=0.0):
        super().__init__()
        self.replacement = replacement
        self.isnull = isnull

    def transform(self, X, **transform_kwargs):
        X = X.copy()
        mask = self.isnull(X)
        X[mask] = self.replacement
        return X


class NullIndicator(BaseTransformer):
    """Indicate whether values are null or not"""

    def transform(self, X, **tranform_kwargs):
        return pd.isnull(X).astype(int)



import pathlib
import unittest
from unittest.mock import Mock

from ballet.project import relative_to_contrib
from ballet.validation.project_structure.checks import (
    IsAdditionCheck, IsPythonSourceCheck, ModuleNameCheck,
    RelativeNameDepthCheck, SubpackageNameCheck, WithinContribCheck)

from .util import make_mock_project


class DiffCheckTest(unittest.TestCase):

    def setUp(self):
        self.contrib_module_path = 'foo/features/contrib'
        self.project = make_mock_project(
            None, None, '', self.contrib_module_path)

    def test_relative_to_contrib(self):
        diff = Mock(b_path='foo/features/contrib/abc.py')

        project = Mock()

        def mock_get(path):
            if path == 'contrib.module_path':
                return self.contrib_module_path
            else:
                raise KeyError
        project.config.get.side_effect = mock_get

        expected = pathlib.Path('abc.py')
        actual = relative_to_contrib(diff, project)

        self.assertEqual(actual, expected)

    def test_is_addition_check(self):
        checker = IsAdditionCheck(self.project)

        mock_diff = Mock(change_type='A')
        self.assertTrue(checker.do_check(mock_diff))

        mock_diff = Mock(change_type='B')
        self.assertFalse(checker.do_check(mock_diff))

    def test_is_python_source_check(self):
        checker = IsPythonSourceCheck(self.project)

        mock_diff = Mock(b_path='foo/features/contrib/user_bob/feature_1.py')
        self.assertTrue(checker.do_check(mock_diff))

        mock_diff = Mock(b_path='foo/features/contrib/user_bob/feature_1.xyz')
        self.assertFalse(checker.do_check(mock_diff))

    def test_within_contrib_check(self):
        checker = WithinContribCheck(self.project)

        mock_diff = Mock(b_path='foo/features/contrib/user_bob/feature_1.py')
        self.assertTrue(checker.do_check(mock_diff))

        mock_diff = Mock(b_path='foo/hack.py')
        self.assertFalse(checker.do_check(mock_diff))

    def test_subpackage_name_check(self):
        checker = SubpackageNameCheck(self.project)

        mock_diff = Mock(b_path='foo/features/contrib/user_bob/feature_1.py')
        self.assertTrue(checker.do_check(mock_diff))

        mock_diff = Mock(b_path='foo/features/contrib/bob/feature_1.py')
        self.assertFalse(checker.do_check(mock_diff))

    def test_feature_module_name_check(self):
        checker = ModuleNameCheck(self.project)

        mock_diff = Mock(b_path='foo/features/contrib/user_bob/feature_1.py')
        self.assertTrue(checker.do_check(mock_diff))

        bad_paths = [
            'foo/features/contrib/user_bob/foo1.py',
            'foo/features/contrib/user_bob/1.py',
            'foo/features/contrib/user_bob/feature_x-1.py',
        ]
        for path in bad_paths:
            mock_diff = Mock(b_path=path)
            self.assertFalse(checker.do_check(mock_diff))

    def test_relative_name_depth_check(self):
        checker = RelativeNameDepthCheck(self.project)

        mock_diff = Mock(b_path='foo/features/contrib/user_bob/feature_1.py')
        self.assertTrue(checker.do_check(mock_diff))

        mock_diff = Mock(
            b_path='foo/features/contrib/user_bob/a/b/c/d/feature_1.py')
        self.assertFalse(checker.do_check(mock_diff))
